/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package algo;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.UUID;
import java.util.zip.GZIPOutputStream;

import jsc.distributions.Distribution;
import jsc.distributions.Normal;
import jsc.distributions.Poisson;
import jsc.distributions.Uniform;
import logging.ILoggingService;
import logging.StaticLogger;
import messages.GroupMessage;
import messages.Message;
import messages.Query;
import network.NetworkNodeGen;
import network.NodeSystem;
import network.RTT;
import remote.IRemoteControllableNetwork;
import remote.RemoteController;
import services.dss.DSSStats;
import services.dss.DynamicSchedulingService;
import services.dss.PoolStatusManager;
import services.dss.profile.ProfileGenerator;
import services.dss.profile.ResourceProfile;
import services.dss.scheduler.pool.Pool;
import support.FloodJob;
import support.INetworkNode;
import support.IStar;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;
import syncutil.AdjacencyMap;
import syncutil.Clustering;
import syncutil.DegreeDistribution;
import syncutil.Girth;
import syncutil.ICompute;
import syncutil.ParallelCompute;
import syncutil.SequentialCompute;
import util.LogPrintStream;
import util.SWFLoader;
import ants.AntSystem;
import ants.BalancingAnt;
import ants.ISSQueryAnt;
import ants.QueryAnt;
import bebop.client.Client;
import cli.CLIOptions;
import cli.DoubleOption;
import cli.FloatOption;
import cli.ICLIOption;
import cli.ICLIOptions;
import cli.InformationHeader;
import cli.IntegerOption;
import cli.LongOption;
import cli.StringOption;
import cli.SwitchOption;

public abstract class Algorithm implements IRemoteControllableNetwork {

	protected ICLIOptions cli;
	//public static IStar xstar;
	
	private ICLIOption<String> logfile = null;
	
	private LogPrintStream logSOut = null;
	
	public static ProfileGenerator jobp;
	
	public static SWFLoader swfdata = null;
	
	public static Algorithm instance;

	public class ProgressStar implements IStar {

		private char[] chars = { '|', '/', '-', '\\' };
		private int current = 0;
		private PrintStream output;
		
		public ProgressStar(PrintStream p) {
			output = p;
		}

		public void show() {
			output.print(get());
		}

		public void hide() {
			output.print('\b');
		}

		private char get() {
			current = (++current) % 4;
			return chars[current];
		}

		public synchronized void progress() {
			hide();
			show();
		}
	}

	public class ProgressStarFake implements IStar {

		public void show() {
		}

		public void hide() {
		}

		public synchronized void progress() {
		}

	}

	protected long step = 0;
	private int radius = 0;
	private int diameter = 0;

	private ICLIOption<Boolean> printcli;
	private ICLIOption<Boolean> cliHelp;
	private ICLIOption<Boolean> cliInfo;
	private ICLIOption<Boolean> bebop;
	private ICLIOption<Integer> bebopport;

	private int tcount = 0;
	private int cpus = 0;

	private ICompute pcomputer;
	
	private Girth girthcomputer = new Girth();
	
	private DegreeDistribution degreecomputer = new DegreeDistribution();
	
	private Clustering clustercomputer = new Clustering();

	private String formatTime(long millis) {
		long hours = millis / 3600000l;
		if (hours > 100) {
			return DateFormat.getDateTimeInstance(DateFormat.LONG,
					DateFormat.LONG).format(millis);
		}
		long minutes = (millis - hours * 3600000) / 60000;
		long seconds = (millis - (minutes * 60000 + hours * 3600000)) / 1000;
		long ms = millis - (minutes * 60000 + hours * 3600000 + seconds * 1000);
		return String.format("%dh %dm %ds %dms", hours, minutes, seconds, ms);
	}
	
	private String formatTime2(long millis) {
		long hours = millis / 3600000l;
		if (hours > 1000) {
			return DateFormat.getDateTimeInstance(DateFormat.LONG,
					DateFormat.LONG).format(millis);
		}
		long minutes = (millis - hours * 3600000) / 60000;
		long seconds = (millis - (minutes * 60000 + hours * 3600000)) / 1000;
		long ms = millis - (minutes * 60000 + hours * 3600000 + seconds * 1000);
		return String.format("%dh %dm %ds %dms", hours, minutes, seconds, ms);
	}

	protected void parseCLI() {
		
		//PrintStream baseOutput = System.out;
		
		List<String> names = new ArrayList<String>();
		for (java.lang.reflect.Field x : Parameters.class.getFields()) {
			names.add("class support.Parameters."+x.getName());
		}
		
		if (!cli.parse() || cliHelp.getValue()) {
			if (cliInfo.getValue())
				InformationHeader.print();
			else
				cli.printUsage();
			System.exit(1);
		}

		
		if (logfile.getValue() != null) {
			try {
				logSOut = new LogPrintStream(new FileOutputStream(logfile.getValue()), System.out);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}
			System.setOut(logSOut); 
		}
		
		if (printcli.getValue())
			System.out.println(cli.toString());
		
		cli.apply(names);
		
		if (names.size() != 0) {
			for (String n : names) {
				System.out.format("(WARN) %s not set from command line\n", n);
			}
		}
		
		// connect nodes at a distance >= b+1, b = 2D-2, i.e. connect if distance is >= 2D-1
		Parameters.connthreshold = 2 * Parameters.D - 1; // connect if greater or equal
		
		// break cycles of length < a, i.e. disconnect if neighbors are at distance < a -2 
		// a = b+2, disconnect if neighbors at distance < b+2-2 = b = 2D-2, i.e brak if <= 2D-3
		//Parameters.deconnthreshold = Parameters.D - 1; // disconnect if less or equal
		Parameters.deconnthreshold = 2* Parameters.D - 3; // disconnect if less or equal

		Randomizer.init(Parameters.RANDOM_SEED);

		if (Parameters.DEBUG) {
			StaticLogger.addOutput(ILoggingService.ALL, System.out);
		}

		if (bebop.getValue()) {
			new RemoteController(bebopport.getValue(), this);
		}

		/*if (Parameters.REAL_STAR)
			Algorithm.star = new ProgressStar(baseOutput);
		else
			Algorithm.star = new ProgressStarFake();*/

		cpus = Runtime.getRuntime().availableProcessors();
		tcount = Math.max((int) (cpus * Parameters.CPUFACTOR),1);
		
		if (Parameters.CPUCOUNT > 0) {
			tcount = Parameters.CPUCOUNT;
		}

		AntSystem.init(1000, tcount);
		NodeSystem.init(1000, tcount);
		NetworkSystem.init(1000, tcount);
		NetworkSystem.rcn = this;

		if (Parameters.SEQ_COMPUTE)
			pcomputer = new SequentialCompute();
		else
			pcomputer = new ParallelCompute(tcount);

		AntSystem.initRunners();
		NodeSystem.initRunners();
		
		if (Parameters.SWF_FILE != null) {
			swfdata = new SWFLoader(Parameters.SWF_FILE);
		}
		// cli.checkCoverage();
	}

	protected void setupCLI(String[] args) {
		cli = new CLIOptions(args);

		printcli = new SwitchOption(null, "printcli", "Outputs commandline");
		cli.addOption(printcli);
		
		logfile = new StringOption(null, "log", "Log file", null,
				false);
		cli.addOption(logfile);
		

		/* General options */
		cliHelp = new SwitchOption("h", "help", "Show this help");
		cli.addOption(cliHelp);
		cliInfo = new SwitchOption(null, "info",
				"Show copyright info and redistribution license");
		cli.addOption(cliInfo);

		/* Simulation parameters */
		cli.addOption(new IntegerOption(null, "maxiter", "Maximum iterations",
				2500, false).linkTo(Parameters.class, "MAX_ITER"));
		cli.addOption(new FloatOption(null, "cpufactor",
				"Threads / Cores ratio", 1f, false).linkTo(Parameters.class,
				"CPUFACTOR"));
		
		cli.addOption(new IntegerOption(null, "cpucount",
				"Cores number", -1, false).linkTo(Parameters.class,
				"CPUCOUNT"));

		cli.addOption(new LongOption(null, "seed", "Random Seed", -1, false)
				.linkTo(Parameters.class, "RANDOM_SEED"));
		cli.addOption(new LongOption(null, "minitermillis",
				"Minimum milliseconds per iteration", -1, false).linkTo(
				Parameters.class, "MIN_ITER_MILLIS"));

		/* Computation */
		cli.addOption(new IntegerOption(null, "cfrom",
				"Compute from iteration x", 0, false).linkTo(Parameters.class,
				"COMPUTE_FROM"));
		cli.addOption(new IntegerOption(null, "cinterval", "Compute interval",
				200, false).linkTo(Parameters.class, "COMPUTE_INTERVAL"));
		cli.addOption(new SwitchOption(null, "star", "Show progress star")
				.linkTo(Parameters.class, "REAL_STAR"));
		cli
				.addOption(new SwitchOption(null, "logcompute",
						"Log compute interval").linkTo(Parameters.class,
						"LOG_COMPUTE"));
		cli.addOption(new FloatOption(null, "logscale", "Log scale", 1.5f,
				false).linkTo(Parameters.class, "LOG_SCALE"));
		cli.addOption(new IntegerOption(null, "dinterval",
				"Distance interval (only for static!)", -1, false).linkTo(
				Parameters.class, "DISTANCE_INTERVAL"));
		cli.addOption(new SwitchOption(null, "nonlimitsearch",
				"Do not limit search (nodes will keep forwarding)").linkTo(
				Parameters.class, "NONLIMIT_SEARCH"));
		cli.addOption(new SwitchOption(null, "coverage",
				"Coverage test instead of resource discovery").linkTo(
				Parameters.class, "COVERAGE"));

		/* Dynamic network */
		cli.addOption(new IntegerOption(null, "remint", "Remove interval", -1,
				false).linkTo(Parameters.class, "REMOVE_NODE_INTERVAL_MS"));
		cli.addOption(new IntegerOption(null, "leaveint", "Leave interval", -1,
				false).linkTo(Parameters.class, "LEAVE_NODE_INTERVAL_MS"));
		cli.addOption(new IntegerOption(null, "addint", "Add node interval",
				-1, false).linkTo(Parameters.class, "ADD_NODE_INTERVAL_MS"));
		cli.addOption(new IntegerOption(null, "searchint", "Search interval",
				-1, false).linkTo(Parameters.class, "SEARCH_NODE_INTERVAL_MS"));
		cli.addOption(new IntegerOption(null, "addstart",
				"When to start adding nodes", -1, false).linkTo(
				Parameters.class, "ADD_NODE_START_MS"));

		cli.addOption(new IntegerOption(null, "remstart",
				"When to start removing nodes", -1, false).linkTo(
				Parameters.class, "REMOVE_NODE_START_MS"));

		cli.addOption(new IntegerOption(null, "leavestart",
				"When to start making nodes leave", -1, false).linkTo(
				Parameters.class, "LEAVE_NODE_START_MS"));

		cli.addOption(new IntegerOption(null, "searchstart",
				"When to start searching nodes", -1, false).linkTo(
				Parameters.class, "SEARCH_NODE_START_MS"));

		cli.addOption(new IntegerOption(null, "searchcount",
				"When to start searching nodes", -1, false).linkTo(
				Parameters.class, "SEARCH_NODE_COUNT"));

		cli.addOption(new IntegerOption(null, "limitffw",
				"Search limit forwarding", -1, false).linkTo(Parameters.class,
				"SEARCH_LIMIT_FFW"));

		cli.addOption(new IntegerOption(null, "limitttl", "Search limit TTL",
				-1, false).linkTo(Parameters.class, "SEARCH_LIMIT_TTL"));

		cli.addOption(new IntegerOption(null, "addstop",
				"When to stop adding nodes", -1, false).linkTo(
				Parameters.class, "ADD_NODE_STOP_MS"));

		cli.addOption(new IntegerOption(null, "remstop",
				"When to stop removing nodes", -1, false).linkTo(
				Parameters.class, "REMOVE_NODE_STOP_MS"));

		cli.addOption(new IntegerOption(null, "searchstop",
				"When to stop searching nodes", -1, false).linkTo(
				Parameters.class, "SEARCH_NODE_STOP_MS"));

		cli.addOption(new IntegerOption(null, "leavestop",
				"When to stop making nodes leave", -1, false).linkTo(
				Parameters.class, "LEAVE_NODE_STOP_MS"));
		cli.addOption(new SwitchOption(null, "addtorandom",
				"Add to random (instead of first node)").linkTo(
				Parameters.class, "ADD_NODE_RANDOM_HOOK"));
		cli.addOption(new IntegerOption(null, "remcount",
				"How many nodes to remove", -1, false).linkTo(Parameters.class,
				"REM_COUNT"));
		cli.addOption(new IntegerOption(null, "addcount",
				"How many nodes to add", -1, false).linkTo(Parameters.class,
				"ADD_COUNT"));
		cli.addOption(new IntegerOption(null, "leavecount",
				"How many nodes to leave", -1, false).linkTo(Parameters.class,
				"LEAVE_COUNT"));
		cli.addOption(new IntegerOption(null, "csample",
				"Compute sample subset", -1, false).linkTo(Parameters.class,
				"SAMPLE_COMPUTE"));

		cli.addOption(new IntegerOption(null, "shiftint", "Shift interval",
				100, false).linkTo(Parameters.class, "SHIFT_INTERVAL"));

		cli.addOption(new SwitchOption(null, "dss",
				"Enable dynamic scheduling service experiment").linkTo(
				Parameters.class, "DSS"));

		cli.addOption(new SwitchOption(null, "girth", "Compute girth")
		.linkTo(Parameters.class, "GIRTH"));
		
		cli.addOption(new SwitchOption(null, "degree", "Compute degree distribution")
		.linkTo(Parameters.class, "DEGREE"));
		
		cli.addOption(new SwitchOption(null, "smartgrid", "Enable smartgrid prox")
		.linkTo(Parameters.class, "SMARTGRID_PROXY"));
		
		cli.addOption(new SwitchOption(null, "cluster", "Compute cluster")
		.linkTo(Parameters.class, "CLUSTERING"));
		
		/* Debugging */
		cli.addOption(new SwitchOption(null, "debug", "Enable debug output")
				.linkTo(Parameters.class, "DEBUG"));
		cli.addOption(new SwitchOption(null, "notify",
				"Enable notification output")
				.linkTo(Parameters.class, "NOTIFY"));

		/* Fail resilience */
		cli.addOption(new IntegerOption(null, "ettl",
				"Emergency connection TTL", 0, false).linkTo(Parameters.class,
				"EMERGENCY_TTL"));

		cli.addOption(new IntegerOption(null, "lttl", "Leave connection TTL",
				0, false).linkTo(Parameters.class, "LEAVE_LINK_TTL"));

		cli.addOption(new IntegerOption(null, "cretry",
				"Connection retry interval", 2000, false).linkTo(
				Parameters.class, "CONNECTION_RETRY_INTERVAL_MS"));
		
		cli.addOption(new IntegerOption(null, "mthreadprobe",
				"Itarations for multithread probing", 1000, false).linkTo(
				Parameters.class, "DO_MULTITHREAD_PROBE"));

		cli.addOption(new IntegerOption(null, "forcettl",
				"Force accept on TTL excess steps", 10, false).linkTo(
				Parameters.class, "FORCE_ACCEPT_ON_TTL"));
		

		cli.addOption(new SwitchOption(null, "spider",
				"Spider web crash recovery").linkTo(Parameters.class,
				"SPIDER_WEB_CRASH_RECOVERY"));
		
		cli.addOption(new SwitchOption(null, "norandomcache",
		"Disable random number generator cache").linkTo(Parameters.class,
		"NO_RANDOM_CACHE"));

		/* Network problems */
		cli.addOption(new FloatOption(null, "slow", "Ant slowdown probability",
				0.0f, false).linkTo(Parameters.class, "PACKET_SLOW_PROB"));

		cli.addOption(new FloatOption(null, "loss", "Ant loss probability",
				0.0f, false).linkTo(Parameters.class, "PACKET_LOSS_PROB"));

		cli.addOption(new IntegerOption(null, "startloss",
				"Start loosing at iteration", -1, false).linkTo(
				Parameters.class, "PACKET_LOSS_START_MS"));

		cli.addOption(new IntegerOption(null, "stoploss",
				"Stop loosing at iteration", -1, false).linkTo(
				Parameters.class, "PACKET_LOSS_STOP_MS"));

		cli.addOption(new IntegerOption(null, "startslow",
				"Start slowdown at iteration", -1, false).linkTo(
				Parameters.class, "PACKET_SLOW_START_MS"));

		cli.addOption(new IntegerOption(null, "stopslow",
				"Stop slowdown at iteration", -1, false).linkTo(
				Parameters.class, "PACKET_SLOW_STOP_MS"));

		/* Input / Output */
		cli.addOption(new StringOption(null, "initial",
				"Initial graph filename", null, true).linkTo(Parameters.class,
				"INITIAL"));

		cli.addOption(new StringOption(null, "prefix", "Output prefix", null,
				false).linkTo(Parameters.class, "PREFIX"));

		cli.addOption(new IntegerOption("f", "fileinterval",
				"File output interval", 10, false).linkTo(Parameters.class,
				"INTERVAL"));

		cli.addOption(new IntegerOption(null, "screeninterval",
				"Screen output interval", 20, false).linkTo(Parameters.class,
				"SCREEN_INTERVAL"));

		cli.addOption(new StringOption(null, "underlying",
				"Underlying network", null, false).linkTo(Parameters.class,
				"UNDERLYING_NETWORK"));

		/* Alpha table */
		cli.addOption(new IntegerOption(null, "alphamaxsize",
				"Alpha table maximum size", 28, false).linkTo(Parameters.class,
				"ALPHA_MAX_SIZE"));
		cli.addOption(new IntegerOption(null, "reqinterval",
				"Alpha table request interval per step", 300, false).linkTo(
				Parameters.class, "REQUEST_MIN_INTERVAL_MS"));

		cli.addOption(new LongOption(null, "alphamaxage",
				"Alpha table maximum age", 900000l, false).linkTo(Parameters.class,
				"ALPHA_MAX_AGE_MS"));

		/* DAnt */
		cli.addOption(new IntegerOption(null, "lv", "Ant vector length", 15,
				false).linkTo(Parameters.class, "ANT_LEN"));

		cli.addOption(new FloatOption(null, "kappa", "Random walk probability",
				0.5f, false)
				.linkTo(Parameters.class, "RANDOM_WALK_PROBABILITY"));

		/* UserLinkAnt */
		cli.addOption(new IntegerOption(null, "ulinkttl", "User link TTL", 20,
				false).linkTo(Parameters.class, "USER_LINK_TTL"));

		/* Evaluation options */
		cli.addOption(new IntegerOption(null, "omegac",
				"Update interval omega connection", 1000, false).linkTo(
				Parameters.class, "UPDATE_INTERVAL_OMEGA_CON_MS"));

		cli.addOption(new IntegerOption(null, "omegad",
				"Update interval omega disconnection", 1000, false).linkTo(
				Parameters.class, "UPDATE_INTERVAL_OMEGA_DIS_MS"));

		cli.addOption(new IntegerOption(null, "d", "D Parameter", 5, false)
				.linkTo(Parameters.class, "D"));

		cli.addOption(new IntegerOption(null, "k", "K Parameter", 1, false)
				.linkTo(Parameters.class, "K"));

		/* Pheromone */
		cli.addOption(new DoubleOption(null, "delta",
				"Default pheromone increment", 0.1f, false).linkTo(
				Parameters.class, "DEF_INCREMENT"));

		cli.addOption(new DoubleOption(null, "initpheromone",
				"Initial pheromone", 1.0f, false).linkTo(Parameters.class,
				"INITIAL_CONCENTRATION"));

		/* Specific pheromone details */
		cli.addOption(new LongOption(null, "gammadecayiter",
				"Iterations for decaying gamma from default to minimum", 30000, false).linkTo(Parameters.class,
				"GAMMA_DECAY_MS"));
		
		/* Specific pheromone details */
		cli.addOption(new LongOption(null, "betadecayiter",
				"Iterations for decaying beta from default to minimum", 60000, false).linkTo(Parameters.class,
				"BETA_DECAY_MS"));

		cli.addOption(new IntegerOption(null, "rttfactor", "RTT Factor (ms per iteration)", 1,
				false).linkTo(Parameters.class, "MS_PER_ITER"));

		/* Node options */
		cli.addOption(new IntegerOption(null, "m", "Max total connections", 8,
				false).linkTo(Parameters.class, "MAX_CONNECTIONS"));

		cli.addOption(new IntegerOption(null, "mo",
				"Max connections from the algorithm", 6, false).linkTo(
				Parameters.class, "MAX_OPTI_CONNECTIONS"));

		/* Ant respawn options */
		cli.addOption(new IntegerOption(null, "antage", "Ant maximum age", 25,
				false).linkTo(Parameters.class, "ANT_MAX_AGE"));

		cli.addOption(new DoubleOption(null, "birthprob",
				"Ant birth probability", 0.05f, false).linkTo(Parameters.class,
				"ANT_BIRTH_PROBABILITY"));

		cli.addOption(new IntegerOption(null, "iota", "Ant respawn interval",
				100000, false).linkTo(Parameters.class, "IOTA_MS"));

		cli.addOption(new SwitchOption(null, "completelcc",
				"Complete LCC calculation").linkTo(Parameters.class,
				"COMPLETE_NS_LCC"));

		cli.addOption(new IntegerOption(null, "nscache", "Newscast cache size",
				20, false).linkTo(Parameters.class, "NS_CACHE_SIZE"));

		cli.addOption(new IntegerOption(null, "nsinterval",
				"Newscast merge interval", 600000, false).linkTo(Parameters.class,
				"NS_MERGE_INTERVAL_MS"));
		
		cli.addOption(new IntegerOption(null, "gnutellaping",
				"Gnutella ping", 30000, false).linkTo(Parameters.class,
				"GNUTELLA_PING_MS"));
		
		cli.addOption(new IntegerOption(null, "gnutellaneighbors",
				"Gnutella neighbors", 10, false).linkTo(Parameters.class,
				"GNUTELLA_NEIGHBORS"));
		
		cli.addOption(new IntegerOption(null, "gnutellattl",
				"Gnutella TTL", 7, false).linkTo(Parameters.class,
				"GNUTELLA_TTL"));
		// see http://wiki.limewire.org/index.php?title=Program.Gnutella.Paper
		cli.addOption(new IntegerOption(null, "gnutellafw",
				"Gnutella FW", 4, false).linkTo(Parameters.class,
				"GNUTELLA_FW"));
		
		cli.addOption(new IntegerOption(null, "traildecaystep",
				"Pheromone trail decay step", 10, false).linkTo(Parameters.class,
				"TRAIL_DECAY_STEP"));
		
		cli.addOption(new LongOption(null, "noderesolution",
				"Node resolution time", 1000, false).linkTo(Parameters.class,
				"NODE_RESOLUTION_INTERVAL_MS"));
		

		/* Remote Control */
		bebop = new SwitchOption(null, "bebop", "Enable BeBop remote control");
		cli.addOption(bebop);
		bebopport = new IntegerOption(null, "bebopport", "BeBop server port",
				34567, false);
		cli.addOption(bebopport);

		cli.addOption(new SwitchOption(null, "seqcompute",
				"Enable sequential compute instead of parallel").linkTo(
				Parameters.class, "SEQ_COMPUTE"));

		cli
				.addOption(new SwitchOption(null, "balancing",
						"Balancing algorithm").linkTo(Parameters.class,
						"BALANCING"));

		/* ISS Proactive caching */

		cli.addOption(new IntegerOption(null, "isscachesize", "ISS Cache size",
				10, false).linkTo(Parameters.class, "ISS_CACHE_SIZE"));

		cli.addOption(new IntegerOption(null, "issmergeint",
				"ISS Merge interval", 250000, false).linkTo(Parameters.class,
				"ISS_MERGE_INTERVAL_MS"));
		cli.addOption(new IntegerOption(null, "issproactiveint",
				"ISS Proactive query interval", 2500000, false).linkTo(
				Parameters.class, "ISS_PROACTIVE_QUERY_INTERVAL_MS"));
		cli.addOption(new IntegerOption(null, "issproactiveffw",
				"ISS Proactive query forwards", 3, false).linkTo(
				Parameters.class, "ISS_PROACTIVE_QUERY_FFW"));
		cli.addOption(new IntegerOption(null, "issproactivettl",
				"ISS Proactive query TTL", 5, false).linkTo(Parameters.class,
				"ISS_PROACTIVE_QUERY_TTL"));
		cli.addOption(new SwitchOption(null, "issenabled", "Enable ISS")
				.linkTo(Parameters.class, "ISS_ENABLED"));
		
		cli
				.addOption(new SwitchOption(null, "adaptiveiss",
						"Enable adaptive ISS").linkTo(Parameters.class,
						"ADAPTIVE_ISS"));
		cli
				.addOption(new SwitchOption(null, "disjointiss",
						"Visit in the ISS overlay don't count for restricted flooding in base overlay")
						.linkTo(Parameters.class, "ISS_DISJOINT_OVERLAYS"));
		cli.addOption(new IntegerOption(null, "issttl", "ISS Query TTL", 3,
				false).linkTo(Parameters.class, "ISS_QUERY_TTL"));
		cli.addOption(new IntegerOption(null, "issffw",
				"ISS Query forwarding limit", 10, false).linkTo(
				Parameters.class, "ISS_QUERY_FFW"));
		cli.addOption(new LongOption(null, "mergestart",
				"ISS Merge start iteration", -1, false).linkTo(
				Parameters.class, "ISS_MERGE_START_MS"));
		
		cli.addOption(new LongOption(null, "pquerystart",
				"ISS Proactive Query start iteration", -1, false).linkTo(
				Parameters.class, "ISS_PROACTIVE_QUERY_START_MS"));
		cli.addOption(new SwitchOption(null, "alwaysforwardleafs",
				"Always forward to leafs (regardless of forwarding limit)")
				.linkTo(Parameters.class, "ALWAYS_FORWARD_LEAFS"));
		
		cli.addOption(new SwitchOption(null, "nonstopsearch",
				"Don't stop search when item found").linkTo(Parameters.class,
				"NONSTOP_SEARCH"));
		
		cli.addOption(new SwitchOption(null, "detaildiameter",
				"Detail paths longer than target diameter").linkTo(
				Parameters.class, "DETAIL_DIAMETER"));
		
		cli.addOption(new SwitchOption(null, "querypathtracking",
				"Enables query path tracking to reduce re-transmission")
				.linkTo(Parameters.class, "QUERY_PATH_TRACKING"));
		
		cli.addOption(new SwitchOption(null, "rttenabled",
				"Enforce RTT during migrations").linkTo(Parameters.class,
				"RTT_MIGRATION"));
		
		
		cli.addOption(new SwitchOption(null, "dss", "Enable dynamic scheduling service experiment").linkTo(Parameters.class, "DSS"));

		cli.addOption(new IntegerOption(null, "dssjobminmatches", "Job generator minimum matches", 2, false).linkTo(Parameters.class, "DSS_JOB_MIN_MATCHES"));

		cli.addOption(new DoubleOption(null, "dssdrift", "DSS Run time Estimation error (percentage)",
				0.1f, false).linkTo(Parameters.class, "DSS_DRIFT"));
		
		cli.addOption(new SwitchOption(null, "dssbaddrift", "Bad drift only").linkTo(Parameters.class, "BAD_DRIFT_ONLY"));
		
		cli.addOption(new IntegerOption(null, "dssmaxretries", "Submit maximum retries", -1, false).linkTo(Parameters.class, "DSS_MAX_RETRIES"));
		
		cli.addOption(new IntegerOption(null, "dssinformttl", "DSS Inform message ttl", 6, false).linkTo(Parameters.class, "DSS_INFORM_TTL"));
		
		cli.addOption(new IntegerOption(null, "dssinformffw", "DSS Inform message ffw", 3, false).linkTo(Parameters.class, "DSS_INFORM_FFW"));
		
		cli.addOption(new IntegerOption(null, "dssquerytracksize", "DSS Query track size", 2048, false).linkTo(Parameters.class, "DSS_QUERY_TRACK_SIZE"));
		
		cli.addOption(new SwitchOption(null, "dssvisualscheduler", "Enable dynamic scheduling service visualization").linkTo(Parameters.class, "DSS_VISUAL_SCHEDULER"));
		
		cli.addOption(new LongOption(null, "dssacceptthreshold", "DSS Accept threshold", 180000, false).linkTo(Parameters.class, "INFORM_ACCEPT_THRESHOLD_MS"));
		
		cli.addOption(new LongOption(null, "dssacceptinterval", "DSS Accept interval (in ms)", 50000, false).linkTo(Parameters.class, "ACCEPT_OFFER_WAIT_MS"));
		
		cli.addOption(new LongOption(null, "dssinforminterval", "DSS Inform interval (in ms)", 120000, false).linkTo(Parameters.class, "INFORM_INTERVAL_MS"));
		
		cli.addOption(new IntegerOption(null, "dssinformlastn", "DSS Inform last n", 2, false).linkTo(Parameters.class, "INFORM_LAST_N"));
		
		cli.addOption(new LongOption(null, "dssmaxjobtime", "DSS Max job time (in ms)", 14400000, false).linkTo(Parameters.class, "DSS_MAX_JOBTIME_MS"));
		cli.addOption(new LongOption(null, "dssminjobtime", "DSS Min job time (in ms)", 3600000, false).linkTo(Parameters.class, "DSS_MIN_JOBTIME_MS"));
		
		cli.addOption(new DoubleOption(null, "dssfifoprob", "DSS FIFO Scheduler probability",
				0.5f, false).linkTo(Parameters.class, "FIFO_PROBABILITY"));
		
		cli.addOption(new SwitchOption(null, "dssedf", "Use EDF scheduler for DSS").linkTo( Parameters.class, "DSS_EDF"));
		cli.addOption(new SwitchOption(null, "dssar", "Use Advance Reservation (AR) scheduler for DSS").linkTo( Parameters.class, "DSS_AR"));
		cli.addOption(new SwitchOption(null, "dssarfair", "Use Fair Advance Reservation (FAR) scheduler for DSS").linkTo( Parameters.class, "DSS_AR_FAIR"));
		cli.addOption(new SwitchOption(null, "dssarfairpool", "Use Fair Advance Reservation (FAR) with job pooling scheduler for DSS").linkTo( Parameters.class, "DSS_AR_FAIR_POOL"));
		cli.addOption(new IntegerOption(null, "dssarfairpoolmaxsize", "Use Fair Advance Reservation (FAR) with job pooling scheduler for DSS; max pool size", 4, false).linkTo( Parameters.class, "DSS_AR_FAIR_MAX_POOL_SIZE"));
		cli.addOption(new LongOption(null, "dssmaxdeadline", "DSS Max deadline", 36000000l, false).linkTo(Parameters.class, "DSS_MAX_DEADLINE_MS"));
		cli.addOption(new LongOption(null, "dssmindeadline", "DSS Min deadline", 18000000l, false).linkTo(Parameters.class, "DSS_MIN_DEADLINE_MS"));
		
		cli.addOption(new SwitchOption(null, "skiparchmatch", "Skip matching of architecture")
		.linkTo(Parameters.class, "SKIP_ARCH_MATCH"));
		cli.addOption(new SwitchOption(null, "skiposmatch", "Skip matching of operating system")
		.linkTo(Parameters.class, "SKIP_OS_MATCH"));
		cli.addOption(new SwitchOption(null, "skipvermatch", "Skip matching of version")
		.linkTo(Parameters.class, "SKIP_VER_MATCH"));
		cli.addOption(new SwitchOption(null, "skipmemmatch", "Skip matching of available memory")
		.linkTo(Parameters.class, "SKIP_MEM_MATCH"));
		cli.addOption(new SwitchOption(null, "skipdiskmatch", "Skip matching of available disk space")
		.linkTo(Parameters.class, "SKIP_DISK_MATCH"));
		
		cli.addOption(new SwitchOption(null, "nocpuspeed", "Do not use different cpu speeds")
		.linkTo(Parameters.class, "NO_CPUSPEED"));
		
		cli.addOption(new SwitchOption(null, "rttgen", "Generate RTT if not available")
		.linkTo(Parameters.class, "RTT_GENERATOR"));
		
		
		cli.addOption(new SwitchOption(null, "profilecache", "Profile cache")
		.linkTo(Parameters.class, "PROFILE_CACHE"));
		cli.addOption(new IntegerOption(null, "profilecachesize", "Profile cache size", 5, false).linkTo( Parameters.class, "REPLICATION_CACHE_SIZE"));
		cli.addOption(new LongOption(null, "repantint", "Replicant interval (ms)", 600000l, false).linkTo( Parameters.class, "REPLICATION_ANT_INTERVAL_MS"));
		
		
		cli.addOption(new StringOption(null, "swf", "SWF File", null,
				false).linkTo(Parameters.class, "SWF_FILE"));
		cli.addOption(new IntegerOption(null, "swfscale", "SWF Scale", 1,
				false).linkTo(Parameters.class, "SWF_SCALE"));
		
		cli.addOption(new SwitchOption(null, "uniformprofiles", "Create resource profiles with uniform probabilities")
		.linkTo(Parameters.class, "UNIFORM_PROFILES"));
		
		cli.addOption(new SwitchOption(null, "onehop", "One hop replication")
		.linkTo(Parameters.class, "ONE_HOP_REPLICATION"));
		cli.addOption(new LongOption(null, "profilecacheage", "Profile cache age", 1200000l, false).linkTo( Parameters.class, "CACHE_TIME_MS"));
		
		cli.addOption(new StringOption(null, "dssbroker", "DSS Broker (job submission node)", null,
				false).linkTo(Parameters.class, "DSS_BROKER"));
		
		cli
		.addOption(new SwitchOption(null, "profilesimilarity",
				"Enable profile based similarity").linkTo(Parameters.class,
				"PROFILE_SIMILARITY"));
		
		cli.addOption(new IntegerOption(null, "qcsize",
				"Query cache size", 128, false).linkTo(
				Parameters.class, "QUERY_CACHE_SIZE"));
		
		cli.addOption(new SwitchOption(null, "gnustats", "Gnutella stats")
		.linkTo(Parameters.class, "GNUSTATS"));
		
		cli.addOption(new SwitchOption(null, "connerror",
		"Connection errors statistics").linkTo(
		Parameters.class, "CONN_ERROR"));
		
		cli.addOption(new SwitchOption(null, "nodisconnect",
		"Disable disconnections").linkTo(
		Parameters.class, "NO_DISCONNECTION"));
		
		
		
	}

	protected void printConfiguration() {
		Parameters.print();
		System.out.format("Using %d parallel threads on %d CPUs\n", tcount,
				cpus);
		System.out.format("---------------------------------------------\n");

	}

	public static int checkProfile(ResourceProfile p) {
		int count = 0;
		for (INetworkNode n : NetworkSystem.getNetworkNodes()) {
			if (n.getProfile().matches(p))
				count++;
		}
		return count;
	}

	// got from http://forums.sun.com/thread.jspa?messageID=9573656
	/*protected int poisson(double c) {
		int x = 0;
		double t = 0.0;
		for (;;) {
			t = t - Math.log(Randomizer.pop()) / c;
			if (t > 1.0)
				break;
			x = x + 1;
		}
		return x;
	}*/

	public void attachNode(String node, String hook) {
		addNode(node, hook);
	}

	public void detachNode(String node) {
		if (NetworkSystem.existNode(node))
			leaveNode(node);
	}

	protected abstract void generateAnts();
	
	public abstract void generateAnt(String node);

	protected void init() {
		instance = this;
		Parameters.MS_PER_ITER_DOUBLE = new Double(Parameters.MS_PER_ITER);
		System.out.println("Milliseconds (double) per iteration: "+Parameters.MS_PER_ITER_DOUBLE);
		RTT.initRTT(false, Parameters.UNDERLYING_NETWORK);
		NetworkNodeGen.init();
	}

	protected void execute() {
		long startTime = System.currentTimeMillis();
		long lastSimTime = System.currentTimeMillis();
		long lastIter = 0;
		
		Client.init();
		
		List<String> lockedNodes = new ArrayList<String>();
		generateAnts();
		Stats.resetAntMigrations();

		int totalpoolsize = 0;
		int totalfound = 0;
		int totalhits = 0;
		int totalhitsOnISS = 0;

		int totalqueryHops = 0;
		int totalsuccess = 0;
		int totalisshops = 0;
		int totalqueries = 0;
		
		double av = (Parameters.DSS_MAX_JOBTIME_MS - Parameters.DSS_MIN_JOBTIME_MS) / 2;
		double stdev = av * 0.5;
		av = Parameters.DSS_MIN_JOBTIME_MS + av;
		Distribution normal = new Normal(av, stdev);

		jobp = new ProfileGenerator(Parameters.UNIFORM_PROFILES, new Uniform(), normal);

		boolean computed = false;
		float avpl = 0;
		int edges = 0;
		long removeAt = Parameters.REMOVE_NODE_START_MS;
		long leaveAt = Parameters.LEAVE_NODE_START_MS;
		long addAt = Parameters.ADD_NODE_START_MS;
		long searchAt = Parameters.SEARCH_NODE_START_MS;
		double lossprob = Parameters.PACKET_LOSS_PROB;
		double slowprob = Parameters.PACKET_SLOW_PROB;

		int totmerges = 0;
		int totissmerges = 0;
		int totissmergescost = 0;
		int totprohops = 0;
		int totprohits = 0;
		int totproreplies = 0;
		int totprostarted = 0;
		int totisspings = 0;
		
		Poisson poisson_LEAVE_NODE_INTERVAL_MS = new Poisson(Math.max(Parameters.LEAVE_NODE_INTERVAL_MS,1));
		Poisson poisson_ADD_NODE_INTERVAL_MS = new Poisson(Math.max(Parameters.ADD_NODE_INTERVAL_MS,1));
		Poisson poisson_REMOVE_NODE_INTERVAL_MS = new Poisson(Math.max(Parameters.REMOVE_NODE_INTERVAL_MS,1));

		if (Parameters.PACKET_LOSS_START_MS >= 0) {
			Parameters.PACKET_LOSS_PROB = 0;
		}
		if (Parameters.PACKET_SLOW_START_MS >= 0) {
			Parameters.PACKET_SLOW_PROB = 0;
		}
		
		Parameters.NODE_RESOLUTION_INTERVAL_ITER = RTT.mstoiter(Parameters.NODE_RESOLUTION_INTERVAL_MS);

		AdjacencyMap a = new AdjacencyMap();
		AdjacencyMap b = new AdjacencyMap();
		int mapdist = -1;
		a.update();
		
		//long startIterTime = System.currentTimeMillis();
		//long waitFor = -1;
		
		//long nextGeneration = 0;
		
		long stepms = 0;

		for (;;) {
			
			stepms = (long) (step * Parameters.MS_PER_ITER_DOUBLE);
			
			/*
			waitFor = Parameters.MIN_ITER_MILLIS
					- Math.abs(System.currentTimeMillis() - startIterTime);
			if (waitFor > 0) {
				try {
					Thread.sleep(waitFor);
				} catch (InterruptedException e) {
				}
			}*/
			//startIterTime = System.currentTimeMillis();

			if ((Parameters.PACKET_LOSS_STOP_MS >= 0)
					&& (stepms >= Parameters.PACKET_LOSS_STOP_MS)) {
				Parameters.PACKET_LOSS_PROB = 0;
			} else if ((Parameters.PACKET_LOSS_START_MS >= 0)
					&& (stepms >= Parameters.PACKET_LOSS_START_MS)) {
				Parameters.PACKET_LOSS_PROB = lossprob;
			}

			if ((Parameters.PACKET_SLOW_STOP_MS >= 0)
					&& (stepms >= Parameters.PACKET_SLOW_STOP_MS)) {
				Parameters.PACKET_SLOW_PROB = 0;
			} else if ((Parameters.PACKET_SLOW_START_MS >= 0)
					&& (stepms >= Parameters.PACKET_SLOW_START_MS)) {
				Parameters.PACKET_SLOW_PROB = slowprob;
			}

			if ((Parameters.DISTANCE_INTERVAL > 0)
					&& (step % Parameters.DISTANCE_INTERVAL == 0)) {
				try {
					b.update();
					mapdist = a.distanceCopyFrom(b);
				} catch (Exception e) {
					mapdist = -1;
				}
			} else {
				mapdist = -1;
			}

			computed = false;
			
			
			if (step % Parameters.DO_MULTITHREAD_PROBE == 0) {
				if ((java.lang.management.ManagementFactory.getOperatingSystemMXBean().getSystemLoadAverage() / cpus) > 0.95) {
					Parameters.DO_MULTITHREAD = false;
				} else {
					Parameters.DO_MULTITHREAD = true;
				}
			}
			
			this.algorithmStep();
			
			/* Run ants */
			//long startt = System.currentTimeMillis();
			//AntSystem.multithreadStep();
			AntSystem.step();
			//System.out.println("Ant step: "+ (System.currentTimeMillis() - startt));

			//startt = System.currentTimeMillis();
			NodeSystem.multithreadStep();
			//NetworkSystem.step();
			//System.out.println("Node step: "+ (System.currentTimeMillis() - startt));
			
			//star.hide();

			/* Export data */
			if ((Parameters.INTERVAL > 0) && (step % Parameters.INTERVAL == 0)) {
				if (Parameters.PREFIX != null) {
					exportGraph(String.format("%s-%d.yaml", Parameters.PREFIX,
							step));
				}
			}

			/* Generate ants */
			/*if (stepms >= nextGeneration) {
				generateAnts();
				nextGeneration += Parameters.IOTA_MS;
			}*/

			/* Export data */
			if ((Parameters.MAX_ITER > 0) && (step > Parameters.MAX_ITER)) {
				if (Parameters.PREFIX != null) {
					exportGraph(String.format("%s-%d.yaml", Parameters.PREFIX,
							step));
				}
				closeSOut();
				System.exit(0);
			}

			/* Leaving of a node */
			if ((leaveAt > 0) && (stepms >= leaveAt)) {
				if ((Parameters.LEAVE_NODE_STOP_MS < 0)
						|| ((Parameters.LEAVE_NODE_STOP_MS > 0) && (stepms < Parameters.LEAVE_NODE_STOP_MS))) {
					leaveAt += poisson_LEAVE_NODE_INTERVAL_MS.random() + 1;
					for (int x = 0; x < Parameters.LEAVE_COUNT; x++) {
						if (NetworkSystem.getNodeCount() == 0) {
							System.out.println("No more nodes");
							closeSOut();
							System.exit(0);
						}
						String no = NetworkSystem.getReallyRandomNode();
						while (lockedNodes.contains(no)) {
							no = NetworkSystem.getReallyRandomNode();
						}
						leaveNode(no);
					}
				}
			}

			/* Removing a node */
			if ((removeAt > 0) && (stepms >= removeAt)) {
				if ((Parameters.REMOVE_NODE_STOP_MS < 0)
						|| ((Parameters.REMOVE_NODE_STOP_MS > 0) && (stepms < Parameters.REMOVE_NODE_STOP_MS))) {
					removeAt += poisson_REMOVE_NODE_INTERVAL_MS.random() + 1;
					for (int x = 0; x < Parameters.REM_COUNT; x++) {
						if (NetworkSystem.getNodeCount() == 0) {
							System.out.println("No more nodes");
							closeSOut();
							System.exit(0);
						}
						String no = NetworkSystem.getReallyRandomNode();
						while (lockedNodes.contains(no)) {
							no = NetworkSystem.getReallyRandomNode();
						}
						killNode(no);
					}
				}
			}

			if (Parameters.DSS
					&& (stepms >= searchAt)
					&& ((Parameters.SEARCH_NODE_STOP_MS < 0) || (searchAt < Parameters.SEARCH_NODE_STOP_MS))) {
				
				INetworkNode d = null;
				
				if (Parameters.DSS_BROKER != null) {
					d = NetworkSystem.getNode(Parameters.DSS_BROKER);
				} else {
					d = NetworkSystem.getNode(NetworkSystem
						.getReallyRandomNode());
				}
				
				addRandomJob(jobp, d);
				
				
				searchAt += Parameters.SEARCH_NODE_INTERVAL_MS;
			}

			/* Searching for a node */
			if ((searchAt > 0) && !Parameters.DSS && (stepms >= searchAt)) {
				if ((Parameters.SEARCH_NODE_STOP_MS < 0)
						|| ((Parameters.SEARCH_NODE_STOP_MS > 0) && (stepms < Parameters.SEARCH_NODE_STOP_MS))) {

					int poolsize = Stats.searchPoolSize();
					int found = Stats.searchFound();
					int hits = Stats.searchHits();

					float searchFoundRate = 0f;
					if (poolsize > 0) {
						searchFoundRate = new Float(found).floatValue()
								/ new Float(poolsize).floatValue();
					}
					int queryHops = Stats.searchQueryHops();
					int success = Stats.searchSuccess();

					totalpoolsize += poolsize;
					totalsuccess += success;
					totalfound += found;
					totalhits += hits;
					totalisshops += Stats.isshops;
					int hitsoniss = Stats.searchHitsOnISS();
					totalhitsOnISS += hitsoniss;

					float totalsearchFoundRate = new Float(totalfound)
							.floatValue()
							/ new Float(totalpoolsize).floatValue() * 100f;
					totalqueryHops += queryHops;
					float totalsearchSuccessRate = new Float(totalsuccess)
							.floatValue()
							/ new Float(totalqueries).floatValue() * 100f;
					float searchSuccessRate = new Float(success).floatValue()
							/ new Float(Parameters.SEARCH_NODE_COUNT)
									.floatValue() * 100f;

					float issContribution = new Float(totalhitsOnISS)
							.floatValue()
							/ new Float(totalhits).floatValue() * 100f;
					int averagehops = 0;
					int averageisshops = 0;
					if (totalqueries > 0) {
						averagehops = totalqueryHops / totalqueries;
						averageisshops = totalisshops / totalqueries;
					}
					System.out
							.format(
									";; %d QS=%d\tTTL=%d/FFWL=%d\tS=%d/PS=%d/F=%d/HT=%d(%d)/FR=%3.2f\tQH=%d(%d)  \tTPS=%d/TF=%d/THT=%d(%d)/TFR=%3.2f(%3.2f)/TQH=%d(%d)\tS=%d/%d/%3.2f/%3.2f/%d(%d)\n",
									step, Parameters.SEARCH_NODE_COUNT,
									Parameters.SEARCH_LIMIT_TTL,
									Parameters.SEARCH_LIMIT_FFW, success,
									poolsize, found, hits, hitsoniss,
									searchFoundRate, queryHops, Stats.isshops,
									totalpoolsize, totalfound, totalhits,
									totalhitsOnISS, totalsearchFoundRate,
									issContribution, totalqueryHops,
									totalisshops, totalsuccess, totalqueries,
									searchSuccessRate, totalsearchSuccessRate,
									averagehops, averageisshops);

					if (AntSystem.getAntCount(QueryAnt.class) != 0) {
						System.out
								.println("Warning! Still some ants from previous search");
					}

					lockedNodes.clear();
					Stats.clearSearchJobs();
					for (int i = 0; i < Parameters.SEARCH_NODE_COUNT; i++) {
						String start = NetworkSystem.getReallyRandomNode();
						
						Message m = null;
						
						if (Parameters.COVERAGE) {
							m = new GroupMessage(UUID.randomUUID().toString(),start);
						} else {
							ResourceProfile profile = jobp.generate(UUID.randomUUID()
									.toString());
							while (checkProfile(profile) < Parameters.DSS_JOB_MIN_MATCHES) {
								profile = jobp.generate(UUID.randomUUID().toString());
							}
							m = new Query(start, profile);
						}
						FloodJob job = new FloodJob(start, m);
						lockedNodes.add(start);

						Stats.addSearchJob(job);
						job.start();
						totalqueries++;
					}

					searchAt += Parameters.SEARCH_NODE_INTERVAL_MS;
				}
			}

			/* Adding a new node */
			if ((addAt > 0) && (stepms >= addAt)) {
				if ((Parameters.ADD_NODE_STOP_MS < 0)
						|| ((Parameters.ADD_NODE_STOP_MS > 0) && (stepms < Parameters.ADD_NODE_STOP_MS))) {
					addAt += poisson_ADD_NODE_INTERVAL_MS.random() + 1;
					for (int x = 0; x < Parameters.ADD_COUNT; x++) {
						String id = NetworkNodeGen.grab();
						String hook;
						if (Parameters.ADD_NODE_RANDOM_HOOK)
							hook = NetworkSystem.getQuasiRandomNode();
						else
							hook = NetworkSystem.getFirstNode();
						addNode(id, hook);
					}
				}
			}

			/* Compute */
			if ((step == -1)
					|| (!Parameters.LOG_COMPUTE
							&& (Parameters.COMPUTE_INTERVAL > 0)
							&& (step > Parameters.COMPUTE_FROM) && (step
							% Parameters.COMPUTE_INTERVAL == 0))) {
				Stats.largestpartition = 0;
				Stats.smallestpartition = 10000000;
				pcomputer.computeBFS();
				diameter = pcomputer.getDiameter();
				radius = pcomputer.getRadius();
				avpl = pcomputer.getAveragePathLength();
				computed = true;
				
				if (Parameters.GIRTH)
					girthcomputer.computeGirth();
				
				if (Parameters.DEGREE)
					degreecomputer.computeDegreeDistribution();
				
				if (Parameters.CLUSTERING)
					clustercomputer.computeClustering();
			}

			/* Display stats */
			if (computed
					|| ((Parameters.SCREEN_INTERVAL > 0) && (step
							% Parameters.SCREEN_INTERVAL == 0))) {
				if (!computed)
					edges = edges + Stats.add - Stats.remove;
				step++;
				int nodescount = NetworkSystem.getNodeCount();
				edges = NetworkSystem.getEdgeCount();
				int overloaded = NetworkSystem.getOverloadedCount();
				String computedString = computed ? "!" : "?";
				int remaining = nodescount * Parameters.MAX_CONNECTIONS - edges;
				System.out
						.format(
								"\n%d\tN=%d(%d/%d/+%d/-%d/~%d)\tE=%d(%d/+%d/-%d)\t%sD/R/a=%d/%d/%3.2f\tS/L/X/D=%d/%d/%d/%d\tNST=%d/P=%d\n",
								step - 1,
								nodescount, Stats.largestpartition,
								Stats.smallestpartition, Stats.newnodes,
								Stats.leavenodes, Stats.remnodes, edges,
								remaining, Stats.add, Stats.remove,
								computedString, diameter, radius, avpl,
								Stats.slow, Stats.loss, overloaded, mapdist,
								Stats.nstokens, Stats.pings);
				System.out.println(AntSystem.getAntTypes());
				totmerges += Stats.merges;
				totissmerges += Stats.issmerges;
				totissmergescost += Stats.issmergescost;
				totprohops += Stats.prohops;
				totprohits += Stats.prohits;
				totproreplies += Stats.proreplies;
				totprostarted += Stats.prostarted;
				totisspings += Stats.isspings;

				System.out.format("\t@ %d", step);

				for (String c : Stats.antMigrations.keySet()) {
					System.out.format("/%s>%d>%d", c.substring(5),
							Stats.antMigrations.get(c),
							Stats.antMigrationsTokens.get(c));
					if (c.substring(5).equals("DiscoveryAnt")) {
						System.out.format(">%d", Stats.nontokens);
					}
				}

				System.out.println();

				System.out
						.format(
								"\t| MRG=%d/ISS(MRG=%d/MRGTK=%d/PHOPS=%d/PHITS=%d/PREPLY=%d/PQS=%s/PINGS=%d)\tTMRG=%d/ISS(TMRG=%d/TMRGTK=%d/TPHOPS=%d/TPHITS=%d/TPREPLY=%d/TPQS=%s/TPINGS=%d)\n",
								Stats.merges, Stats.issmerges,
								Stats.issmergescost, Stats.prohops,
								Stats.prohits, Stats.proreplies,
								Stats.prostarted, Stats.isspings, totmerges,
								totissmerges, totissmergescost, totprohops,
								totprohits, totproreplies, totprostarted,
								totisspings);

				if (Parameters.BALANCING) {
					System.out.format("\t§ %d\tBA=%d\tB=%d\tLV=%3.2f\n",
							step - 1,
							AntSystem.getAntCount(BalancingAnt.class),
							Stats.balancing, Stats.loadVariance());
				}

				System.out.format("\t§ avMINT=%d\tavPINT=%d\tavCSIZE=%3.2f\n",
						NetworkSystem.getAverageMergeInterval(), NetworkSystem
								.getAverageProactiveInterval(), NetworkSystem
								.getAverageCacheSize());
				
				if (Parameters.PROFILE_CACHE) {
					System.out.format("\t+ RS=%d\tRH=%d\tP=%d\tRHits=%d\tTRS=%d\tTRH=%d\tTP=%d\tTRHits=%d\n", Stats.replicantstarted, Stats.replicanthops, Stats.publish, Stats.replicanthit, Stats.treplicantstarted, Stats.treplicanthops, Stats.tpublish, Stats.treplicanthit);
				}

				if (Parameters.RTT_MIGRATION)
					System.out.format("\t¢ %s (sim ends at %s)\n", formatTime((step - 1)
							* Parameters.MS_PER_ITER), formatTime2(Parameters.MAX_ITER * Parameters.MS_PER_ITER));
				
				long msperiter = 0;
				try {
					msperiter = (System.currentTimeMillis() - lastSimTime) / (step - lastIter);
				} catch (Exception e) {
				}
				
				System.out.format("\t(TIME) ms/Iter = %s, started = %s, elapsed = %s, ETC = %s\n", formatTime2(msperiter), formatTime(startTime), formatTime2(System.currentTimeMillis() - startTime), formatTime2((Parameters.MAX_ITER - step) * msperiter));
				//lastSimTime = System.currentTimeMillis();
				//lastIter = step;
				
				if (Parameters.GIRTH)
					girthcomputer.printGirth();
				
				if (Parameters.DEGREE)
					degreecomputer.printDegree();
				
				if (Parameters.CLUSTERING)
					clustercomputer.printClustering();	
				
				if (Parameters.GNUSTATS) {
					System.out.format("\tGnutella: PING=%d, PONG=%d, CONNECT=%d, TPING=%d, TPONG=%d, TCONNECT=%d\n", Stats.gnuping, Stats.gnupong, Stats.gnuconnect, Stats.tgnuping, Stats.tgnupong, Stats.tgnuconnect);
				}
				
				double avconn = 0;
				if (Stats.totconnect != 0)
					avconn = Stats.totconntime / Stats.totconnect;
				
				double avdisconn = 0;
				if (Stats.totdisconnect != 0)
					avdisconn = Stats.totdisconntime / Stats.totdisconnect;
				
				System.out.format("\tAverage ev conn: %3.2f ms (%3f), Average ev disconn: %3.2f ms (%3f) \n", avconn, Stats.totconnect, avdisconn, Stats.totdisconnect);
				
				System.out.println("\t(CD) C="+Stats.connects+" D="+Stats.disconnects);
					
				if (Parameters.CONN_ERROR) {
					double perc = 0;
					if (Stats.triedconn > 0) {
						perc = (double) Stats.wrongconn/ (double) Stats.triedconn*100.0;
					}
					double operc = 0;
					if (Stats.otriedconn > 0) {
						operc = (double) Stats.owrongconn/ (double) Stats.otriedconn*100.0;
					}
					System.out.println("\t(CERR) IC="+Stats.initiatedconn+" TC="+Stats.triedconn+" WC="+Stats.wrongconn+ " WCperc="+perc+" OIC="+Stats.oinitiatedconn+" OTC="+Stats.otriedconn+" OWC="+Stats.owrongconn+ " OWCperc="+operc);
				}
			
				printDSSStats(step);

				printCustomStats();

				System.out.println();

				Stats.reset();
				Stats.resetNodes();
			} else
				step++;

			NetworkSystem.tick();
		}
	}

	protected abstract void algorithmStep();

	private void closeSOut() {
		if (logSOut != null) {
			logSOut.close();
		}
	}

	public static void addRandomJob(ProfileGenerator jobp, INetworkNode d) {
		if (swfdata != null) {
			ResourceProfile job = swfdata.getJob(NetworkSystem.getGlobalTimeMs());
			while (job != null) {
				d.getDSSService().addJob(job);
				job = swfdata.getJob(NetworkSystem.getGlobalTimeMs());
			}
		} else if (Parameters.DSS_AR_FAIR_POOL) {
			ResourceProfile profile = jobp.generate(UUID.randomUUID()
					.toString());
			int count = (int) (Randomizer.pop() * Parameters.DSS_AR_FAIR_MAX_POOL_SIZE) + 1;
			while (checkProfile(profile) < count) {
				profile = jobp.generate(UUID.randomUUID().toString());
			}
			String poolid = UUID.randomUUID().toString();
			List<ResourceProfile> jobpool = jobp.generatePool(profile, poolid, count);
			//List<String> jobs = new ArrayList<String>();
			Pool p = new Pool(d, poolid, profile);
			for (ResourceProfile j : jobpool) {
				// System.out.println("[SUBMIT] Node "+d.getIdentifier()+" receives new job "+profile);
				d.getDSSService().addJob(j);
				p.addJob(j.id);
			}
			((PoolStatusManager) d.getDSSService().getJobStatusManager()).addPool(p);
		} else {
			ResourceProfile profile = jobp.generate(UUID.randomUUID()
					.toString());
			while (checkProfile(profile) < Parameters.DSS_JOB_MIN_MATCHES) {
				profile = jobp.generate(UUID.randomUUID().toString());
			}
			// System.out.println("[SUBMIT] Node "+d.getIdentifier()+" receives new job "+profile);
			d.getDSSService().addJob(profile);
		}
	}

	public List<String> getNodeList() {
		List<String> ls = new ArrayList<String>();
		ls.addAll(NetworkSystem.getNodes());
		return ls;
	}

	public void removeNode(String node) {
		if (NetworkSystem.existNode(node))
			killNode(node);
	}
	
	protected void exportGraph(String path) {
		List<String> nodes = new ArrayList<String>();
		nodes.addAll(NetworkSystem.getNodes());
		Collections.sort(nodes, new Comparator<String>() {

			public int compare(String o1, String o2) {
				Integer o1i = Integer.parseInt(o1);
				Integer o2i = Integer.parseInt(o2);
				return o1i.compareTo(o2i);
			}

		});
		
		FileOutputStream s = null;
		GZIPOutputStream fos = null;
		try {
			s = new FileOutputStream(path+".gz");
			fos = new GZIPOutputStream(s);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		
		
		try {
			fos
					.write("!!python/object:networkx.graph.Graph\nadj:\n"
							.getBytes());
			for (String n : nodes) {
				INetworkNode node = NetworkSystem.getNode(n);
				fos.write(String.format("  %s:\n", n).getBytes());
				List<String> neighbors = new ArrayList<String>();
				// ********************************************************
				neighbors.addAll(node.getNeighbors());
				// ********************************************************
				Collections.sort(neighbors, new Comparator<String>() {

					public int compare(String o1, String o2) {
						Integer o1i = Integer.parseInt(o1);
						Integer o2i = Integer.parseInt(o2);
						return o1i.compareTo(o2i);
					}

				});
				for (String neigh : neighbors) {
					// if (NetworkSystem.getNode(neigh).isNeighbor(n)) {
					fos
							.write(String.format("    %s: null\n", neigh)
									.getBytes());
					// }
				}
			}
			fos.write("name: Subgraph of ()\n".getBytes());
			fos.finish();
			fos.close();
			s.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	protected abstract void addNode(String id, String hook);

	protected final synchronized void killNode(String id) {
		if (NetworkSystem.existNode(id)) {
			if (Parameters.NOTIFY)
				System.out.format("Killed %s\n", id);
			INetworkNode thenode = NetworkSystem.getNode(id);
			thenode.acquire();
			Stats.remNode();
			NetworkSystem.removeNode(id);
			thenode.release();
		}
	}

	protected final synchronized void leaveNode(String id) {
		if (NetworkSystem.existNode(id)) {
			if (Parameters.NOTIFY)
				System.out.format("Killed %s (cleanly)\n", id);
			INetworkNode thenode = NetworkSystem.getNode(id);
			thenode.acquire();
			thenode.cleanLeave();
			Stats.leaveNode();
			NetworkSystem.removeNode(id);
			thenode.release();
		}
	}

	protected final synchronized void startQuery(String start, Query query,
			int ttl) {
		if (Parameters.ISS_ENABLED) {
			AntSystem.schedule(new ISSQueryAnt(new String(start), null,
					new String(start), query, ttl, false, null));
		} else {
			AntSystem.schedule(new QueryAnt(new String(start), null,
					new String(start), query, ttl, null));
		}
	}

	public final void query(String start, Query query) {
		AntSystem.schedule(new QueryAnt(new String(start), null, new String(
				start), query, Parameters.SEARCH_LIMIT_TTL, null));
	}

	public final synchronized List<String> getNeighbors(String node) {
		List<String> ns = new ArrayList<String>();
		if (NetworkSystem.existNode(node)) {
			INetworkNode thenode = NetworkSystem.getNode(node);
			thenode.acquire();
			ns.addAll(thenode.getNeighbors());
			thenode.release();
		}
		return ns;
	}
	
	public void printDSSStats(long step) {
		if (Parameters.DSS) {
			int totalJobs = 0;
			int totalWaiting = 0;
			long totalScheduleLength = 0;
			int nonEmptySchedules = 0;
			int nodes = NetworkSystem.getNodeCount();
			
			for (String n : NetworkSystem.getNodes()) {
				INetworkNode node = NetworkSystem.getNode(n);
				DynamicSchedulingService thedss = node.getDSSService();
				totalJobs += thedss.getScheduler().getScheduleSize();
				totalWaiting += thedss.getAcceptOfferManager().getAcceptingQueueSize();
				long length = thedss.getScheduler().getScheduleLength();
				
				totalScheduleLength += length;
				if (length != 0)
					nonEmptySchedules++;
			}
			
			System.out.format("\t~ %d, TotNodes %d, TotSchedJobs %d, TotWaiting %d, TotScheduleTime %d, EmptyQueues %d, NonEmptyQueues %d\n" , 
					step, nodes, totalJobs, totalWaiting, totalScheduleLength, nodes-nonEmptySchedules, nonEmptySchedules);
			DSSStats.liveStats();
		}
	}

	public void printCustomStats() {

	}

}
