/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package algo;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import node.CentralizedNode;
import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import util.Dijkstra;
import util.Graph;

public class CentralizedBlatAnt extends Algorithm {

	@Override
	protected void addNode(String id, String hook) {
		// Do nothing
	}

	@Override
	protected void generateAnts() {
		// Do nothing
	}
	
	public void generateAnt(String n) {
		
	}
	
	
	public static void main(String[] args) {
		new CentralizedBlatAnt(args);
	}
	
	public CentralizedBlatAnt(String[] args) {
		setupCLI(args);
		parseCLI();

		System.out.format("---------------------------------------------\n");
		System.out.println("BlatAnt Centralized Implementation");
		System.out.println("(C) 2008 Amos Brocco <amos.brocco@unifr.ch>");
		System.out.format("---------------------------------------------\n");

		printConfiguration();
		init();
		importYaml(Parameters.INITIAL);
		System.out.format("NODES = %d\n", NetworkSystem.getNodeCount());
		System.out.format("---------------------------------------------\n");
		
		this.execute();
	}
	
	private Graph getNetworkGraph() {
		return  NetworkSystem.getGraph(false);
	}
	
	private void importYaml(String path) {

		String line = null;
		String currentNode = null;
		BufferedReader br = null;

		try {
			br = new BufferedReader(new FileReader(path));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}

		try {
			while ((line = br.readLine()) != null) {
				if (line.charAt(0) != ' ') {
					// Skip this
				} else {
					StringTokenizer st = new StringTokenizer(line, ":");
					String id = new String(st.nextToken().trim());
					if (id.equals("datastructure"))
						break;
					if (st.countTokens() > 0) {
						// Neighbor
						CentralizedNode node;
						if (!NetworkSystem.existNode(id)) {
							node = new CentralizedNode(new String(id));
							NetworkSystem.addNode(node);
						} else {
							node = (CentralizedNode) NetworkSystem.getNode(id);
						}
						node.userconnect(new String(currentNode));
						NetworkSystem.getNode(currentNode).userconnect(new String(id));
					} else {
						// New node
						CentralizedNode node;
						if (!NetworkSystem.existNode(id)) {
							node = new CentralizedNode(new String(id));
							NetworkSystem.addNode(node);
						}
						currentNode = id;
					}
				}
			}
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	@Override
	protected void algorithmStep() {
		if (step >= 2) {
			disconnectionStep();
			connectionStep();
		}
	}
	
	
	private void connectionStep() {
		boolean clean = true;
		Graph graph = getNetworkGraph();
		
		List<String> tabuPairs = new ArrayList<String>();
		
		System.out.println("Checking connection rule");
		clean = true;
		graph = getNetworkGraph();
		tabuPairs = new ArrayList<String>();
		
		int dijkstra = 0;
		int connections = 0;
		int rounds = 0;

		// Connection rule
		
		for(;;) {
			for (String n : NetworkSystem.getNodes()) {
				for (String m : NetworkSystem.getNodes()) {
					if (tabuPairs.contains(n+"-"+m)) continue;
					
					INetworkNode nnode = NetworkSystem.getNode(n);
					INetworkNode mnode = NetworkSystem.getNode(m);
					
					//if (nnode.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS) continue;
					//if (mnode.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS) continue;
					
					int distance = Dijkstra.getShortestPath(graph, n, m).size() - 1;
					dijkstra++;
					if (distance >= Parameters.connthreshold) {
						// Connect nodes
						nnode.connect(m);
						graph.addEdge(n, m);
						mnode.connect(n);
						graph.addEdge(m, n);
						tabuPairs.add(n+"-"+m);
						tabuPairs.add(m+"-"+n);
						connections++;
						//System.out.println("\t! "+n+" connecting to "+m);
						//break;
					} else {
						tabuPairs.add(n+"-"+m);
					}
				}
				if (!clean) break;
			}
			if (clean) break;
			clean = false;
		}
		
		System.out.println("Connections:"+connections+ " Djikstra:"+dijkstra+" Rounds:"+rounds);
	}
	
	private void disconnectionStep() {
		
		boolean clean = true;
		Graph graph = getNetworkGraph();
		
		//List<String> tabuPairs = new ArrayList<String>();
		
		System.out.println("Checking disconnection rule");
		
		List<String> nodes = new ArrayList<String>();
		nodes.addAll(NetworkSystem.getNodes());
		
		if (graph.getVertexCount() != NetworkSystem.getNodeCount()) {
			System.out.println("Node count mismatch... something went wrong!");
			System.exit(1);
		}
		
		int dijkstra = 0;
		int disconnections = 0;
		int rounds = 0;
		
		boolean restart = false;
		
		do {
			clean = true;
			rounds++;
			for (String n : nodes) {
				//System.out.println("Processing "+ n);
				Set<String> succ = graph.getSuccessors(n);
				for (String ns1 : succ) {
					for (String ns2 : succ) {
						if (ns1.equals(ns2)) continue;
						//if (tabuPairs.contains(ns1+"-"+ns2)) continue;
						List<String> path = Dijkstra.getShortestPath(graph, ns1, ns2, n);
						dijkstra++;
						int distance = path.size() - 1;
						//System.out.println("(DISC) Checking "+ns1+" -> "+ ns2+ ":"+ path +" dist:"+distance);
						if (distance < 1) continue;
							
						if (distance <= Parameters.deconnthreshold) {
							// Disconnect neighbors
							String candidate = ns2;
							if (NetworkSystem.getNode(ns1).getNeighborsCount() > NetworkSystem.getNode(ns2).getNeighborsCount())
								candidate = ns1;
							NetworkSystem.getNode(n).disconnect(candidate);
							NetworkSystem.getNode(candidate).disconnect(n);
							graph = getNetworkGraph();
							//System.out.println("\t! "+n+" disconnecting "+candidate+" d="+distance+" "+ns1+"<->"+ns2+" on "+path);
							//tabuPairs.add(ns1+"-"+ns2);
							//tabuPairs.add(ns2+"-"+ns1);
							clean = false;
							disconnections++;
							restart = true;
							break;
						} else {
							//tabuPairs.add(ns1+"-"+ns2);
						}
					}
					if (restart) break;
				}
				if (restart) break;
			}
			if (restart) {
				restart = false;
			}
			//System.out.println("Disconnections:"+disconnections+ " Djikstra:"+dijkstra+" Rounds:"+rounds);
		} while(!clean);
		
		
		System.out.println("Disconnections:"+disconnections+ " Djikstra:"+dijkstra+" Rounds:"+rounds);
	}
	
	

}
