/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package alpha;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import logging.StaticLogger;
import node.BlatAntNode;
import support.NetworkSystem;
import support.Parameters;
import support.Stats;
import util.AStar;
import util.BFS;
import util.Graph;
import ants.AntSystem;
import ants.OptimizationLinkAnt;
import ants.UnlinkAnt;
import ants.UpdateNeighborsAnt;

public class Alpha {

	protected String identifier;
	protected BlatAntNode owner;
	protected Table table;
	protected boolean changed = false;
	protected Entry localInfo;

	public Alpha(BlatAntNode o) {
		table = new Table(o, Parameters.ALPHA_MAX_SIZE);
		owner = o;
		identifier = owner.getIdentifier();
		localInfo = new Entry(identifier);
	}
	
	private void neighborsChanged() {
		for (String n : getNeighbors()) {
			AntSystem.schedule(new UpdateNeighborsAnt(new String(identifier), new String(n)));
		}
	}

	public void addNeighbor(String id) {
		/*if (Parameters.ISS_ENABLED || Parameters.ADAPTIVE_ISS)
			owner.getInformationSystemService().countNetworkActivity();*/
		if (!localInfo.isNeighbor(id)) {
			Stats.addLink();
			if (Parameters.NOTIFY)
				System.out.format("Adding %s -> %s\n", identifier, id);
			StaticLogger.log("%s added new neighbor %s", (Object) identifier,
					id);
			localInfo.addNeighbor(id);
			table.stick(table.getIndexOrNew(id));
			neighborsChanged();
		}
	}

	public void clear() {
		table.clear();
		localInfo.clear();
	}

	/**
	 * Evaluates a connection to another node
	 */
	protected void evaluateConnection() {
		
		Stats.evConn();
		
		// /////////////////////////////////////////////////////////
		// PRECONDITION, WE CAN ADD NEW CONNECTIONS
		// /////////////////////////////////////////////////////////
		if ((Parameters.MAX_OPTI_CONNECTIONS > 0)
				&& (getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS))
			return;
		
		// /////////////////////////////////////////////////////////
		// CLEANUP THE ALPHA TABLE
		// /////////////////////////////////////////////////////////
		for (Entry e : table) {
			if ((e == null) || (e.isFree())) continue;
			if ((owner.getTime() - e.getEntryAge()) > Parameters.ALPHA_MAX_AGE_MS) {// Too old
				if (!e.isSticky())
					e.free();
			}
		}

		// /////////////////////////////////////////////////////////
		// GRAPH CONSTRUCTION FROM N_i
		// /////////////////////////////////////////////////////////
		
		long startTime = System.currentTimeMillis();

		Graph temp = new Graph(false);
		temp.construct(identifier, localInfo.getNeighbors(), table);
		temp.removeNonBidirectionalEdges();

		// /////////////////////////////////////////////////////////
		// EVALUATE THE DISTANCES
		// /////////////////////////////////////////////////////////
		// AStar as = new AStar(temp);
		BFS bfs = new BFS(temp);
		
		int max_distance = Parameters.connthreshold - 1;
		String max_distance_node = null;
		Entry max_distance_entry = null;

		for (Entry e : table) {
			if (e == null) // Non valid
				continue;
			if (e.isFree()) // Non valid
				continue;
			if (e.isSticky()) // Neighbor
				continue;
			if (getEstimatedDistance(e.getId()) < Parameters.connthreshold) { // Distance
																			// is
																			// not
																			// sufficient
				continue;
			}
			if (e.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS) { // Don't
																			// test
																			// if
																			// target
																			// is
																			// full
				continue;
			}
			if (owner.getTime() < e.getNextRequestMinTime()) { // If
																		// it's
																		// not
																		// yet
																		// time
																		// to
																		// repeat
																		// the
																		// request
																		// don't
																		// do it
				continue;
			}
			
			
			/* Path to the node */
			int bfslength = bfs.getShortestPathLength(identifier, e.getId());
			
			if (bfslength == -1) {
				temp.removeVertex(e.getId());
				continue;
			}

			e.setComputedDistance(bfslength);

			if (bfslength > max_distance) {
				//System.out.println("Found distance "+bfslength);
				max_distance = bfslength;
				max_distance_node = e.getId();
				max_distance_entry = e;
				e.setNextRequestMinTime(
						owner.getTime() + max_distance
								* Parameters.REQUEST_MIN_INTERVAL_MS);
			}
		}
		
		
	
		if (max_distance >= Parameters.connthreshold) { // >= 2D - 1
			/* Compute the detailed path to the most distant node */
			/*
			 * List<String> pathfwd = as.getShortestPath(identifier, new
			 * String( max_distance_node));
			 */

			// Save the minimum next request timestamp (to avoid sending too
			// much requests)
			// This is computed on the basis of the length to be traveled, and
			// an estimated
			// time per step.
			// table.get(max_distance_node).setNextRequestMinTimestamp(owner.getTime()
			// + pathfwd.size()* Parameters.REQUEST_MIN_INTERVAL);

			// Do not connect here (it slows down convergence) */
			// AntSystem.schedule(AntFactory.getLinkAnt(new String(identifier),
			// new String(max_distance_node), pathfwd));
			max_distance_entry.setDistance(1); // Assume we get accepted
			Stats.initiatedconn();
			AntSystem.schedule(new OptimizationLinkAnt(new String(identifier),
					new String(max_distance_node)));
		}
		
		Stats.connectionTime(System.currentTimeMillis() - startTime);
	}

	/**
	 * Evaluates a disconnection from a neighbor
	 */
	private void evaluateDisconnection() {
		
		if (Parameters.NO_DISCONNECTION) return;
		
		Stats.evDisConn();
		
		// /////////////////////////////////////////////////////////
		// CLEANUP THE ALPHA TABLE
		// /////////////////////////////////////////////////////////
		for (Entry e : table) {
			if ((e == null) || (e.isFree())) continue;
			if ((owner.getTime() - e.getEntryAge()) > Parameters.ALPHA_MAX_AGE_MS) {// Too old
				if (!e.isSticky())
					e.free();
			}
		}

		// /////////////////////////////////////////////////////////
		// GRAPH CONSTRUCTION FROM N_i
		// /////////////////////////////////////////////////////////
		
		long startTime = System.currentTimeMillis();

		Graph temp = new Graph(true);
		temp.constructForDisconnection(identifier, table);

		// Candidate set (neighbors that we will consider for removal)
		Set<String> nprime = new HashSet<String>();
		for (String n : localInfo.getNeighbors()) {
			if (NetworkSystem.isSuperior(identifier, n))
				nprime.add(n);
		}

		// Remove ourselves
		temp.removeVertex(identifier);
		// Remove non bidirectional edges
		temp.removeNonBidirectionalEdges();

		// C is the set of detected cycles such that Rule 2 applies
		List<List<String>> C = new LinkedList<List<String>>();
		// Evaluate distance between neighbor pairs
		AStar as = new AStar(temp);

		BFS bfs = new BFS(temp);

		for (String p : nprime) {
			if (!isActiveNeighbor(p))
				continue;
			for (String q : localInfo.getNeighbors()) {
				if (p.equals(q))
					continue;
				if (!isActiveNeighbor(q))
					continue;
				// Get an initial estimation of the distance
				int bfslength = bfs.getShortestPathLength(p, q);
				if (bfslength > Parameters.deconnthreshold)
					continue;
				if (bfslength < 1)
					continue;

				// Get the shortest path from p to q
				List<String> cyclepath = as.getShortestPath(new String(p),
						new String(q));

				// Distance between neighbors is length - 1
				// Cycle is of size N+1, i.e max distance if cycle is broken:
				// N+1-1 = N
				List<String> cycle = new ArrayList<String>();
				cycle.addAll(cyclepath);
				cycle.add(identifier);
				C.add(cyclepath);
			}
		}

		// Sort C by ascending d
		Collections.sort(C, new Comparator<List<String>>() {
			public int compare(List<String> o1, List<String> o2) {
				Integer a = o1.size();
				return a.compareTo(o2.size());
			}
		});

		// Break up cycles
		Set<String> X = new HashSet<String>();

		for (List<String> cycle : C) {
			String p = new String(cycle.get(0));
			String q = new String(cycle.get(cycle.size() - 1));
			if (!X.contains(p) && !X.contains(q)) {
				if (isActiveNeighbor(p) && isActiveNeighbor(q)) {
					this.owner.disconnect(p); // Free the connection as soon
												// as possible
					AntSystem.schedule(new UnlinkAnt(new String(identifier),
							new String(p)));
				}
				X.add(p);
				X.add(q);
			}
		}
		
		Stats.disconnectionTime(System.currentTimeMillis() - startTime);
	}

	public int getEstimatedDistance(String target) {
		if (target.equals(identifier))
			return 0;

		if (isNeighbor(target))
			return 1;

		Entry e = table.get(target);
		if (e == null)
			return -1;

		int estimated = e.getDistance();
		int cached = e.getComputedDistance();

		if ((estimated == cached) && (estimated == -1))
			return -1;
		if (estimated == -1)
			return cached;
		if (cached == -1)
			return estimated;

		return Math.min(estimated, cached);
	}

	public void getInfosTo(Entry e) {
		e.updateFrom(localInfo);
	}

	public Collection<String> getNeighbors() {
		return localInfo.getNeighbors();
	}

	public List<String> getNeighbors(String target) {
		Entry e = table.get(target);
		if (e != null) {
			return e.getNeighbors();
		} else {
			return new ArrayList<String>();
		}
	}

	public Integer getNeighborsCount() {
		return localInfo.getNeighborsCount();
	}

	public void getNeighborsTo(Set<String> ns) {
		localInfo.getNeighbors(ns);
	}

	public boolean isActiveNeighbor(String id) {
		return localInfo.isNeighbor(id) && table.contains(id)
				&& table.get(id).getNeighbors().contains(identifier);
	}

	public boolean isNeighbor(String id) {
		return localInfo.isNeighbor(id);
	}

	public void onConnect(String target) {
		if (!localInfo.isNeighbor(target)) {
			addNeighbor(new String(target));
			int i = table.getIndexOrNew(new String(target));
			Entry e = table.get(i);
			e.setDistance(1);
		}
	}

	public void onDisconnect(String target) {
		if (localInfo.isNeighbor(target)) {
			Entry e = table.get(new String(target));
			if (e != null) {
				e.unstick();
				e.removeNeighbor(identifier);
				e.setDistance(Parameters.D);
			}
		}
	}

	public void removeNeighbor(String id) {
		/*if (Parameters.ISS_ENABLED || Parameters.ADAPTIVE_ISS)
			owner.getInformationSystemService().countNetworkActivity();*/
		if (localInfo.isNeighbor(id)) {
			Stats.removeLink();
			int i = table.getIndex(id);
			table.unstick(i);
			localInfo.removeNeighbor(id);
			if (Parameters.NOTIFY)
				System.out.format("Removing %s -> %s\n", identifier, id);
			StaticLogger.log("%s removed neighbor %s", (Object) identifier, id);
			neighborsChanged();
		}
	}

	private long lastConnTry = 0; 
	private long lastDisconnTry = 0;
	
	public void step() {

		if (changed) {
			long ctime = owner.getTime();
			
			if ((ctime - lastDisconnTry) >= Parameters.UPDATE_INTERVAL_OMEGA_DIS_MS) {
				//System.out.println("Evaluate disconnection");
				evaluateDisconnection();
				lastDisconnTry = ctime;
				changed = false;
			}
			
			if ((ctime - lastConnTry) >= Parameters.UPDATE_INTERVAL_OMEGA_CON_MS) {
				//System.out.println("Evaluate connection");
				evaluateConnection();
				lastConnTry = ctime;
				changed = false;
			}
		}
	}

	public void updateEntry(Entry e, int d) {
		Entry entry = table.getOrNew(e.getId());
		long ts = entry.getRemoteTime();
		if (e.getRemoteTime() >= ts) {
			if (entry.updateFrom(e))
				changed = true;
			entry.setEntryAge(owner.getTime());
			if ((d == 1) && !isNeighbor(e.getId()))
				return;
			entry.setDistance(d);
		} else {
			//System.out.println("Not updating "+ts + " vs "+e.getRemoteTime());
		}
	}

	public void updateInformation(List<Entry> info) {
		int d = 1;
		for (int i = info.size() - 1; i >= 0; i--) {
			Entry item = info.get(i);
			String id = (String) item.getId();
			// Stop if we come back to ourselves
			if (id.equals(identifier)) {
				break;
			}
			updateEntry(item, d);
			d++;
		}
	}

	public void updateNeighbors(String neighbor, Entry nentry) {
		Entry e = table.get(neighbor);
		if (e != null) {
			long ts = e.getRemoteTime();
	        if (nentry.getRemoteTime() >= ts) {
	        	if (e.updateNeighbors(nentry.getNeighbors())) {
	        		changed = true;
	        	}
	        }
	        e.setRemoteTime(Math.max(ts,nentry.getRemoteTime()));
	        e.setEntryAge(owner.getTime());
		}
	}

	public void activateNeighbor(String neighbor) {
		if (!isNeighbor(neighbor))
			return;
		Entry n = table.get(neighbor);
		if ((n != null) && !n.isNeighbor(identifier))
			n.addNeighbor(new String(identifier));
	}

}
