/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package alpha;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import logging.StaticLogger;
import node.BlatAntSNode;
import support.NetworkSystem;
import support.Parameters;
import support.Stats;
import ants.AntSystem;
import ants.OptimizationLinkAnt;
import ants.UnlinkAnt;
import ants.UpdateNeighborsAnt;

public class SAlpha {

	private String identifier;
	private BlatAntSNode owner;
	private Table table;
	private Entry localInfo;
	private boolean changed = false;

	public SAlpha(BlatAntSNode o) {
		table = new Table(o, Parameters.ALPHA_MAX_SIZE);
		owner = o;
		identifier = owner.getIdentifier();
		localInfo = new Entry(identifier);
	}
	
	private void neighborsChanged() {
		for (String n : getNeighbors()) {
			AntSystem.schedule(new UpdateNeighborsAnt(new String(identifier), new String(n)));
		}
	}

	public void addNeighbor(String id) {
		/*if (Parameters.ISS_ENABLED || Parameters.ADAPTIVE_ISS)
			owner.getInformationSystemService().countNetworkActivity();*/
		if (!localInfo.isNeighbor(id)) {
			Stats.addLink();
			if (Parameters.NOTIFY)
				System.out.format("Adding %s -> %s\n", identifier, id);
			StaticLogger.log("%s added new neighbor %s", (Object) identifier,
					id);
			localInfo.addNeighbor(id);
			table.stick(table.getIndexOrNew(id));
			neighborsChanged();
		}
	}

	public void clear() {
		table.clear();
		localInfo.clear();
	}

	/**
	 * Evaluates a connection to another node
	 */
	private void evaluateConnection() {
		// /////////////////////////////////////////////////////////
		// PRECONDITION, WE CAN ADD NEW CONNECTIONS
		// /////////////////////////////////////////////////////////
		if ((Parameters.MAX_OPTI_CONNECTIONS > 0)
				&& (getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS))
			return;
		
		// /////////////////////////////////////////////////////////
		// CLEANUP THE ALPHA TABLE
		// /////////////////////////////////////////////////////////
		for (Entry e : table) {
			if ((e == null) || (e.isFree())) continue;
			if ((owner.getTime() - e.getEntryAge()) > Parameters.ALPHA_MAX_AGE_MS) {// Too old
				if (!e.isSticky())
					e.free();
			}
		}

		// /////////////////////////////////////////////////////////
		// EVALUATE THE DISTANCES
		// /////////////////////////////////////////////////////////

		int max_distance = 0;
		String max_distance_node = null;
		Entry max_distance_entry = null;

		for (Entry e : table) {
			if (e == null) // Non valid
				continue;
			if (e.isFree()) // Non valid
				continue;
			if (e.isSticky()) // Neighbor
				continue;
			if ((owner.getTime() - e.getEntryAge()) > Parameters.ALPHA_MAX_AGE_MS) {// Too old
				if (!e.isSticky())
					e.free();
				continue;
			}

			int estimated = getEstimatedDistance(e.getId());

			if (estimated < 0)
				continue;
			if (estimated < Parameters.connthreshold) // Distance is not
														// sufficient
				continue;
			if (e.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS) // Don't
																			// test
																			// if
																			// target
																			// is
																			// full
				continue;
			if (owner.getTime() < e.getNextRequestMinTime()) // If
																		// it's
																		// not
																		// yet
																		// time
																		// to
																		// repeat
																		// the
																		// request
																		// don't
																		// do it
				continue;

			if (estimated > max_distance) {
				max_distance = estimated;
				max_distance_node = e.getId();
				max_distance_entry = e;
				e.setNextRequestMinTime(
						owner.getTime() + max_distance
								* Parameters.REQUEST_MIN_INTERVAL_MS);
			}
		}

		if (max_distance >= Parameters.connthreshold) { // >= 2D - 1
			max_distance_entry.setDistance(1); // Assume we get accepted
			Stats.initiatedconn();
			AntSystem.schedule(new OptimizationLinkAnt(new String(identifier),
					new String(max_distance_node)));
			
		}
	}

	/**
	 * Evaluates a disconnection from a neighbor
	 */
	private void evaluateDisconnection(List<Entry> info) {
		
		if (Parameters.NO_DISCONNECTION) return;
		
		// /////////////////////////////////////////////////////////
		// CLEANUP THE ALPHA TABLE
		// /////////////////////////////////////////////////////////
		for (Entry e : table) {
			if ((e == null) || (e.isFree())) continue;
			if ((owner.getTime() - e.getEntryAge()) > Parameters.ALPHA_MAX_AGE_MS) {// Too old
				if (!e.isSticky())
					e.free();
			}
		}

		int d = 0;
		String last = null;
		
		long startTime = System.currentTimeMillis();

		for (Entry e : info) {

			// If we hit ourselves, reset the data
			if (e.getId().equals(identifier)) {
				last = null;
				d = 0;
				continue;
			}

			// If we encounter a node with a superior id, reset
			if (NetworkSystem.isSuperior(e.getId(), identifier)) {
				last = null;
				d = 0;
				continue;
			}

			// If we encounter an active neighbor
			if (isActiveNeighbor(e.getId())) {
				// ... and there was already a neighbor before
				if (last != null) {
					if (last.equals(e.getId())) {
						d = 0;
					} else if (d <= Parameters.deconnthreshold) { // ... >= D-1, ie < D and
																	// the
																	// distance
																	// is less
																	// than the
																	// limit,
																	// disconnect
						int lastNScount = getNeighbors(last).size();
						int thisNScount = getNeighbors(e.getId()).size();
						// Disconnect the neighbor with the highest degree
						// ... so that we help overloaded nodes
						if (lastNScount > thisNScount) {
							this.owner.disconnect(last); // Free the
															// connection as
															// soon as possible
							AntSystem.schedule(new UnlinkAnt(new String(
									identifier), new String(last)));
							last = null;
							d = 0;
							continue; // FIXME: why this was a break?
						} else {
							this.owner.disconnect(e.getId()); // Free the
																// connection as
																// soon as
																// possible
							AntSystem.schedule(new UnlinkAnt(new String(
									identifier), new String(e.getId())));
							last = null;
							d = 0;
							continue; // FIXME: why this was a break?
						}
					} else {
						// ... reset all parameters to this new neighbor
						d = 0;
						last = e.getId();
					}
				} else {
					// we didn't have a last, set it
					d = 0;
					last = e.getId();
				}
			}
			d++;
		}
		
		Stats.disconnectionTime(System.currentTimeMillis() - startTime);
	}

	public int getEstimatedDistance(String target) {
		if (target.equals(identifier))
			return 0;

		if (isNeighbor(target))
			return 1;
		
		for(String t : getNeighbors()) {
			Entry en = table.get(t);
			if (en != null) {
				if (en.isNeighbor(target))
					return 2;
			}
		}
		
		Entry e = table.get(target);
		if (e == null)
			return -1;

		return e.getDistance();
	}

	public void getInfosTo(Entry e) {
		e.updateFrom(localInfo);
	}

	public Collection<String> getNeighbors() {
		return localInfo.getNeighbors();
	}

	public List<String> getNeighbors(String target) {
		Entry e = table.get(target);
		if (e != null) {
			return e.getNeighbors();
		} else {
			return new ArrayList<String>();
		}
	}

	public Integer getNeighborsCount() {
		return localInfo.getNeighborsCount();
	}

	public void getNeighborsTo(Set<String> ns) {
		localInfo.getNeighbors(ns);
	}

	public boolean isActiveNeighbor(String id) {
		return localInfo.isNeighbor(id) && table.contains(id)
				&& table.get(id).getNeighbors().contains(identifier);
	}

	public boolean isNeighbor(String id) {
		return localInfo.isNeighbor(id);
	}

	public void onConnect(String target) {
		if (!localInfo.isNeighbor(target)) {
			addNeighbor(new String(target));
			int i = table.getIndexOrNew(new String(target));
			Entry e = table.get(i);
			e.setDistance(1);
		}
	}

	public void onDisconnect(String target) {
		if (localInfo.isNeighbor(target)) {
			Entry e = table.get(new String(target));
			if (e != null) {
				e.unstick();
				e.removeNeighbor(identifier);
				e.setDistance(Parameters.D);
			}
		}
	}

	public void removeNeighbor(String id) {
		/*if (Parameters.ISS_ENABLED || Parameters.ADAPTIVE_ISS)
			owner.getInformationSystemService().countNetworkActivity();*/
		if (localInfo.isNeighbor(id)) {
			Stats.removeLink();
			int i = table.getIndex(id);
			table.unstick(i);
			localInfo.removeNeighbor(id);
			if (Parameters.NOTIFY)
				System.out.format("Removing %s -> %s\n", identifier, id);
			StaticLogger.log("%s removed neighbor %s", (Object) identifier, id);
			neighborsChanged();
		}
	}

	public void updateEntry(Entry e, int d) {
		Entry entry = table.getOrNew(e.getId());
		entry.updateFrom(e);
		entry.setEntryAge(owner.getTime());
		if ((d == 1) && !isNeighbor(e.getId()))
			return;
		if(entry.setDistance(d)) {
			changed = true;
		}
	}

	public void updateInformation(List<Entry> info) {
		
		long startTime = System.currentTimeMillis();
		
		int distanceFromNeighbor = 2; // We start from the first node after the neighbor
		Map<String, Integer> distances = new HashMap<String, Integer>();

		for (int i = info.size() - 2; i >= 0; i--) {
			Entry item = info.get(i);
			String id = (String) item.getId();
			if (isNeighbor(id)) {
				distanceFromNeighbor = 2;
			} else {
				distances.put(id, distanceFromNeighbor);
				distanceFromNeighbor++;
			}
		}

		distanceFromNeighbor = Parameters.ANT_LEN; // We start at the end of the vector 
												   // (we don't have a neighbor to evaluate, so estimate
												   // to the length of the vector
		
		for (int i = 0; i < info.size() - 2; i++) {
			Entry item = info.get(i);
			String id = (String) item.getId();
			if (isNeighbor(id)) {
				distanceFromNeighbor = 2;
			} else {
				int value = Math.min(distanceFromNeighbor, distances.get(id));
				distances.put(id, value);
				distanceFromNeighbor++;
			}
		}

		for (String k : distances.keySet()) {
			Entry entry = table.getOrNew(new String(k));
			entry.setEntryAge(owner.getTime());
			if (entry.setDistance(distances.get(k))) {
				changed = true;
			}
		}
		
		Stats.connectionTime(System.currentTimeMillis() - startTime);

		evaluateDisconnection(info);
	}
	
	private long lastConnTry = 0; 
	
	public void step() {
		if (changed) {
			long ctime = owner.getTime();
			if ((ctime - lastConnTry) >= Parameters.UPDATE_INTERVAL_OMEGA_CON_MS) {
				//System.out.println("Evaluate connection");
				evaluateConnection();
				lastConnTry = ctime;
				changed = false;
			}
		}
	}

	public void updateNeighbors(String neighbor, Entry nentry) {
		Entry e = table.get(neighbor);
		if (e != null) {
			/*if (Parameters.ISS_ENABLED || Parameters.ADAPTIVE_ISS)
				owner.getInformationSystemService().countNetworkActivity();*/
			long ts = e.getRemoteTime();
	        if (nentry.getRemoteTime() >= ts) {
	        	if (e.updateNeighbors(nentry.getNeighbors())) {
	        		changed = true;
	        	}
	        }
	        e.setRemoteTime(Math.max(ts,nentry.getRemoteTime()));
	        e.setEntryAge(owner.getTime());
		}
	}

	public int getTimestamp() {
		return (int) localInfo.getRemoteTime();
	}

	public void activateNeighbor(String neighbor) {
		if (!isNeighbor(neighbor))
			return;
		Entry n = table.get(neighbor);
		if ((n != null) && !n.isNeighbor(identifier))
			n.addNeighbor(new String(identifier));
	}

}
