/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package ants;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import messages.Message;
import services.IResourceManagementService;
import services.dss.DSSStats;
import support.IBlatAntNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;

public class ISSDSSInformAnt extends Ant {

	private Message query;
	private String sourcetabu;
	private String previoustabu;
	private int ttl;
	private boolean navigateISS = false;
	private List<String> queryPath;
	private long eta;

	public ISSDSSInformAnt(String source, String previous, String target,
			Message query, int ttl, boolean navigateISS, List<String> querypath, long eta) {
		super(target);
		this.query = query;
		this.ttl = ttl;
		this.eta = eta;
		this.sourcetabu = source;
		this.previoustabu = previous;
		this.navigateISS = navigateISS;
		this.queryPath = new ArrayList<String>();
		if (Parameters.QUERY_PATH_TRACKING && !(querypath == null))
			this.queryPath.addAll(querypath);
		setScheduleAfter(NetworkSystem.getRTT(previous, target));
	}

	@Override
	public void antstep() {
		if ((current == null) || !NetworkSystem.isNodeAlive(current_)) {
			// We are on a dead node
			kill();
			return;
		}

		current.acquire();
		
		// Manage beta pheromone
		if (this.previoustabu != null) {
			if (current instanceof IBlatAntNode) {
				((IBlatAntNode) current).activateNeighbor(previoustabu);
			}
			if (current.getBeta() != null) {
				current.getBeta().fill(new String(previoustabu));
			}
		}

		// Check the query

		IResourceManagementService.ProcessingResult result = current.getResourceManagementService().processQuery(query,
				navigateISS);
				
		/* Check if the query matches */
		if (result.equals(IResourceManagementService.ProcessingResult.STOP)) {
			current.release();
			kill();
			return;
		} else if (result.equals(IResourceManagementService.ProcessingResult.MATCH)) {
			current.getDSSService().inform(sourcetabu, query.getProfile().toMap(), eta);
			if (!navigateISS) {
				ttl = Parameters.ISS_QUERY_TTL;
			}
			navigateISS = true;
		}

		// Forward the query

		if (ttl > 0) {
			--ttl; // Count one hop
			List<String> candidates = new ArrayList<String>();

			int limitffw = Parameters.SEARCH_LIMIT_FFW;
			int count = 0;

			if (navigateISS) {
				limitffw = Parameters.ISS_QUERY_FFW;
				candidates.addAll(current.getInformationSystemService()
						.getCacheIdentifiers());
			} else {
				candidates.addAll(current.getNeighbors()); // Added because of (1)
			}

			candidates.removeAll(queryPath);
			
			if (Parameters.ALWAYS_FORWARD_LEAFS) {
				List<String> toSkip = new ArrayList<String>();
				for (String q : candidates) {
					if (current.getNeighborNeighborsSize(q) == 1) {
						// It's a leaf
						if ((!q.equals(sourcetabu)) && (!q.equals(current_)) && (!q.equals(previoustabu))) {
							// TTL set to 0 to prevent further forwarding if our info is wrong
							AntSystem.schedule(new ISSDSSInformAnt(new String(sourcetabu), new String(current_), new String(q), query, 0, navigateISS, queryPath, eta));
							Stats.messageHop(query, current.getIdentifier(), q);
							statmigration();
							if (current.getGamma() != null)
								current.getGamma().fill(q);
						}
						toSkip.add(q);
					}
				}
				candidates.removeAll(toSkip);
			}
			
			if (Parameters.QUERY_PATH_TRACKING)
				queryPath.add(current.getIdentifier());

			/*
			// (1) This part has been disabled... we don't want to go back on the
			// base overlay if there are no neighbors in the upper one
			if (!navigateISS || (candidates.size() == 0)) {
				navigateISS = false;
				candidates.addAll(current.getNeighbors());
			}
			*/

			Collections.shuffle(candidates, Randomizer.getRandom());

			for (String q : candidates) {
				if ((limitffw != -1) && (count >= limitffw))
					break;
				if ((!q.equals(sourcetabu)) && (!q.equals(current_))
						&& (!q.equals(previoustabu))) {
					AntSystem.schedule(new ISSDSSInformAnt(new String(sourcetabu),
							new String(current_), new String(q), query, ttl, navigateISS, queryPath, eta));
					if (navigateISS) {
						Stats.isshop();
					}
					Stats.messageHop(query, current.getIdentifier(), q);
					statmigration();
					if (current.getGamma() != null)
						current.getGamma().fill(q);
					count++;
				}
			}
		}

		current.release();
		kill();
		return;

	}

	@Override
	public void statmigration() {
		DSSStats.informHop();
	}

}
