/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package node;

import gnutella.GnutellaAnt;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;
import ants.AntSystem;

public class GnutellaNode extends Node {

	private List<String> neighbors;
	
	private class BoundedLinkedHashMap extends LinkedHashMap<String,String> {

		private static final long serialVersionUID = 8400552408674580679L;

		@SuppressWarnings("unchecked")
		protected boolean removeEldestEntry(Map.Entry eldest) {
	        return size() > 1024;
	     }
	     
	     public String put(String key, String value) {
			return super.put(key, value);	 
	     }
	}
	
	private ArrayList<String> seenPings = new ArrayList<String>();
	private ArrayList<String> sentPings = new ArrayList<String>();
	private BoundedLinkedHashMap pingSources = new BoundedLinkedHashMap();
	private long nextCleanup = (long) (Randomizer.pop() * Parameters.GNUTELLA_PING_MS * 2);
	
	private void addSeenPing(String uuid) {
		synchronized(seenPings) {
			if (!seenPings.contains(uuid)) {
				seenPings.add(uuid);
				if (seenPings.size() > 1024) {
					seenPings.remove(0);
				}
			}
		}
	}
	
	private void addSentPing(String uuid) {
		synchronized(sentPings) {
			if (!sentPings.contains(uuid)) {
				sentPings.add(uuid);
				if (sentPings.size() > 1024) {
					sentPings.remove(0);
				}
			}
		}
	}
	
	private void addPingSource(String uuid, String origin) {
		synchronized(pingSources) {
			pingSources.put(uuid, origin);
		}
	}

	public GnutellaNode(String id) {
		super(id);
		neighbors = new ArrayList<String>();
		nextCleanup = (long) (getTime() + Parameters.GNUTELLA_PING_MS * Randomizer.pop());
	}

	public void iterstep() {
		
	}

	public void timestep() {
		
		if (getTime() >= nextCleanup) {
			cleanup();
			nextCleanup  = getTime() + Parameters.GNUTELLA_PING_MS;
		}
	}

	public void cleanLeave() {
		return;
	}

	public boolean connect(String target) {
		synchronized(neighbors) {
			if (id.equals(target))
				return false;
			if (neighbors.contains(target))
				return true;
			if (neighbors.size() >= Parameters.GNUTELLA_NEIGHBORS)
				return false;
			else {
				neighbors.add(target);
				return true;
			}
		}
	}

	public boolean disconnect(String target) {
		return false;
	}

	public Collection<String> getNeighbors() {
		Collection<String> ns = new ArrayList<String>(neighbors.size());
		ns.addAll(neighbors);
		return ns;
	}

	public int getNeighborsCount() {
		return neighbors.size();
	}

	public int getNeighborNeighborsSize(String n) {
		return -1;
	}

	public boolean isNeighbor(String source) {
		return neighbors.contains(source);
	}
	
	/** Gnutella protocol 0.4 functions */
	
	private void sendPing() {
		sendPing(UUID.randomUUID().toString(), Parameters.GNUTELLA_TTL, 0, null);
	}
	
	public boolean userconnect(String target) {
		String trytarget = target;
		/*if (neighbors.size() >= Parameters.GNUTELLA_NEIGHBORS)
			return false;*/
		while(true) {
			if (trytarget != null) {
				INetworkNode tnode = NetworkSystem.getNode(trytarget);
				Stats.gnutellaConnect();
				if (tnode != null) {
					tnode.acquire();
					if (tnode.connect(this.getIdentifier())) {
						neighbors.add(trytarget);
						tnode.release();
						return true;
					} else {
						tnode.release();
					}
				}
			}
			trytarget = NetworkSystem.getReallyRandomNode();
		}
	}

	public void onPing(String origin, String uuid, int ttl, int hops,
			String target) {
		if (sentPings.contains(uuid)) {
			return;
		} else {
			
			if (seenPings.contains(uuid)) {
				return;
			} else {
				addSeenPing(uuid);
				sendPong(origin, uuid, Parameters.GNUTELLA_TTL, 0, new String(getIdentifier()));
				
				if (ttl > 1) {
					sendPing(uuid, ttl-1, hops+1, origin);
					addPingSource(uuid, origin);
				}
			}
		}
	}

	private void sendPing(String uuid, int ttl, int hops, String origin) {
		if (origin == null) {
			addSentPing(uuid);
		}
		
		addSeenPing(uuid);
		List<String> candidates = new ArrayList<String>();
		candidates.addAll(neighbors);
		if (origin != null)
			candidates.remove(origin);
		Collections.shuffle(candidates);
		int count = 0;
		for (String target : candidates) {
			if (!target.equals(origin)) {
				Stats.gnutellaPing();
				AntSystem.schedule(new GnutellaAnt(getIdentifier(), new String(target), GnutellaAnt.TYPE.PING,
					uuid, ttl, hops, null));
				count++;
			}
			if ((Parameters.GNUTELLA_FW > 0) && (count >= Parameters.GNUTELLA_FW))
				break;
		}
	}

	private void sendPong(String target, String uuid, int ttl, int hops, String payload) {
		Stats.gnutellaPong();
		AntSystem.schedule(new GnutellaAnt(getIdentifier(), new String(target), GnutellaAnt.TYPE.PONG,
				uuid, ttl, hops, new String(payload)));
	}

	public void onPong(String origin, String uuid, int ttl, int hops,
			String target, String payload) {
		if (sentPings.contains(uuid)) {
			if (neighbors.size() >= Parameters.GNUTELLA_NEIGHBORS)
				return;
			if (!neighbors.contains(payload)) {
				tryConnect(payload);
			}
		} else if (seenPings.contains(uuid)) {
			if (ttl > 1) {
				String route = pingSources.get(uuid);
				if (route != null)
					this.sendPong(route, uuid, ttl-1, hops+1, payload);
			}
		}
	}

	private void tryConnect(String trytarget) {
		if (neighbors.size() >= Parameters.GNUTELLA_NEIGHBORS)
			return;
		INetworkNode tnode = NetworkSystem.getNode(trytarget);
		Stats.gnutellaConnect();
		if (tnode != null) {
			//tnode.acquire();
			if (tnode.connect(this.getIdentifier())) {
				synchronized(neighbors) {
					neighbors.add(trytarget);
				}
				//tnode.release();
				return;
			} else {
				//tnode.release();
			}
		}
	}
	
	private void cleanup() {
		List<String> tbr = new ArrayList<String>();
		for (String n : neighbors) {
			if(!NetworkSystem.ping(n)) {
				tbr.add(n);
			}
		}
		neighbors.removeAll(tbr);
		if (neighbors.size() == 0) {
			userconnect(null);
		} else if (neighbors.size() < Parameters.GNUTELLA_NEIGHBORS){
			sendPing();
		}
	}
	
}
