/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package node;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;

import jsc.distributions.Poisson;

import support.NetworkSystem;
import support.NewscastEntry;
import support.Parameters;
import support.Randomizer;
import support.Stats;

public class NewscastNode extends Node {

	private Object cacheMutex = new Object();
	private NewscastEntry[] cache;
	private int cachesize = 0;

	private long next = (long) (Randomizer.pop() * Parameters.NS_MERGE_INTERVAL_MS * 2);
	private static Poisson poisson_NS_MERGE_INTERVAL_MS = new Poisson(Math.max(Parameters.NS_MERGE_INTERVAL_MS,1));

	public NewscastNode(String id, int cachesize) {
		super(id);
		this.cachesize = cachesize;
		this.cache = new NewscastEntry[cachesize];
	}

	public String getRandomAlivePeer() {
		synchronized(cacheMutex) {
			String peer = null;
			int tries = 30;
			while ((peer == null) && (tries > 0)) {
				int t = (int) (Randomizer.pop() * cachesize);
				if ((cache[t] != null) && (cache[t].ts >= 0)) {
					peer = cache[t].id;
					if (id.equals(peer))
						continue;
					if (!NetworkSystem.ping(peer))
						peer = null;
				}
				tries--;
			}
			return peer;
		}
	}

	public void pushNeighbor(String n) {
		synchronized(cacheMutex) {
			if (Parameters.NOTIFY)
				System.out.format("Adding %s -> %s\n", id, n);
			cache[0] = new NewscastEntry();
			cache[0].ts = getTime();
			cache[0].id = new String(n);
		}
	}

	public void iterstep() {
		
	}
	
	public void timestep() {
		
		if (getTime() >= next) {
			Stats.merge();
			String remote = getRandomAlivePeer();
			if (remote != null) {
				NewscastNode remoteNode = (NewscastNode) NetworkSystem
						.getNode(remote);
				if (remoteNode != null) {
					mergeData(remoteNode);
				}
			}
			next += poisson_NS_MERGE_INTERVAL_MS .random() + 1;
		}
	}

	public void replaceCacheWith(NewscastEntry[] c) {
		synchronized(cacheMutex) {
			cache = c;
		}
	}

	public NewscastEntry[] getCacheClone() {
		synchronized(cacheMutex) {
			return cache.clone();
		}
	}

	public void printcache() {
		for (int i = 0; i < cachesize; i++) {
			if (cache[i] != null)
				System.out.println(cache[i].id + " --> " + cache[i].ts);
			else
				System.out.println("null");
		}
	}

	public static void main(String[] args) {
		NewscastNode x = new NewscastNode("x", Parameters.NS_CACHE_SIZE);
		NewscastNode y = new NewscastNode("y", Parameters.NS_CACHE_SIZE);

		NewscastEntry[] tx = new NewscastEntry[Parameters.NS_CACHE_SIZE];
		for (int i = 0; i < 5; i++) {
			NewscastEntry op = new NewscastEntry();
			op.id = "TX" + i;
			op.ts = i;
			tx[i] = op;
		}
		x.replaceCacheWith(tx);
		System.out.println("--------- X --------");
		x.printcache();

		NewscastEntry[] ty = new NewscastEntry[Parameters.NS_CACHE_SIZE];
		for (int i = 0; i < 10; i++) {
			NewscastEntry op = new NewscastEntry();
			op.id = "TY" + i;
			op.ts = i;
			ty[i] = op;
		}
		y.replaceCacheWith(ty);
		System.out.println("--------- Y --------");
		y.printcache();

		x.step();
		x.mergeData(y);

		System.out.println("--------- X --------");
		x.printcache();
		System.out.println("--------- Y --------");
		y.printcache();

		System.out.println("--------- X --------");
		x.printcache();
		System.out.println("--------- Y --------");
		y.printcache();
	}

	private void mergeData(NewscastNode remote) {
		
		/*if (Parameters.NOTIFY) {
			// Save actual neighbors;
			oldremoteNeighbors = remoteNode.getNeighbors();
			oldlocalNeighbors = getNeighbors();
		}*/
		
		NewscastEntry[] remoteCache = remote.getCacheClone();
		NewscastEntry[] localCache = getCacheClone();
		NewscastEntry[] temp = new NewscastEntry[2 * cachesize];

		Stats.newscastTokens(2 * cachesize);

		for (int i = 0; i < cachesize; i++) {
			if (localCache[i] == null)
				continue;
			boolean done = false;
			for (int j = 0; j < cachesize * 2; j++) {
				if ((temp[j] != null) && (temp[j].id.equals(localCache[i].id))) {
					// There is already an entry
					temp[j].ts = Math.max(temp[j].ts, localCache[i].ts);
					done = true;
				}
			}
			if (done)
				continue;
			temp[i] = new NewscastEntry(localCache[i]);
		}

		for (int i = 0; i < cachesize; i++) {
			if (remoteCache[i] == null)
				continue;
			boolean done = false;
			for (int j = 0; j < cachesize * 2; j++) {
				if ((temp[j] != null) && (temp[j].id.equals(remoteCache[i].id))) {
					// There is already an entry
					temp[j].ts = Math.max(temp[j].ts, remoteCache[i].ts);
					done = true;
				}
			}
			if (done)
				continue;
			temp[cachesize + i] = new NewscastEntry(remoteCache[i]);
		}

		Arrays.sort(temp, new Comparator<NewscastEntry>() {

			public int compare(NewscastEntry arg0, NewscastEntry arg1) {
				if (arg0 == null)
					return 1;
				if (arg1 == null)
					return -1;
				return ((Long) arg1.ts).compareTo((Long) arg0.ts);
			}
		});

		for (int i = 1; i < cachesize; i++) {
			localCache[i] = temp[i - 1];
		}
		
		boolean simulateLoss = false;
		
		if (Randomizer.pop() < Parameters.PACKET_LOSS_PROB) {
			Stats.loss();
			simulateLoss = true;
		}
		
		if (!simulateLoss) {
			for (int i = 1; i < cachesize; i++) {
				remoteCache[i] = temp[i - 1];
			}
		}
		
		long maxtime = Math.max(getTime(), remote.getTime());
		
		localCache[0] = new NewscastEntry();
		localCache[0].ts = new Long(maxtime);
		localCache[0].id = new String(remote.getIdentifier());
		
		if (!simulateLoss) {
			remoteCache[0] = new NewscastEntry();
			remoteCache[0].ts = new Long(maxtime);
			remoteCache[0].id = new String(id);
		}
		
		replaceCacheWith(localCache);
		remote.replaceCacheWith(remoteCache);

		/*if (!compareCaches(remote))
			System.out.println("No equal caches after merge!");*/

		Stats.addLink();

		// System.out.println(id+" has merged with "+remote.getIdentifier());
		// printcache();
		// remote.printcache();
	}

	/*private boolean compareCaches(NewscastNode remote) {
		NewscastEntry[] remoteCache = remote.getCache();
		for (int i = 1; i < cachesize; i++) {
			if ((cache[i] == null) && (remoteCache[i] == null))
				continue;
			if ((cache[i] == null) && !(remoteCache[i] == null))
				return false;
			if (!(cache[i] == null) && (remoteCache[i] == null))
				return false;
			if (!cache[i].id.equals(remoteCache[i].id))
				return false;
			if (cache[i].ts != remoteCache[i].ts)
				return false;
		}
		return true;
	}*/

	public boolean userconnect(String target) {
		// pushNeighbor(target);
		NewscastNode remoteNode = (NewscastNode) NetworkSystem.getNode(target);
		if (remoteNode == null) {
			// System.out.println("Remote node "+target+" does not exist; cannot
			// connect "+id);
			return false;
		} else {
			setTime(remoteNode.getTime());
			mergeData(remoteNode);
			/*
			 * NewscastNode n = (NewscastNode) NetworkSystem.getNode(target);
			 * this.mergeData(n); return false;
			 */
			return true;
		}
	}

	public boolean connect(String target) {
		return false;
	}

	public boolean disconnect(String target) {
		return false;
	}

	public Collection<String> getNeighbors() {
		synchronized(cacheMutex) {
			Collection<String> ns = new ArrayList<String>(cachesize);
			for (int i = 0; i < cachesize; i++) {
				NewscastEntry e = cache[i];
				if (e != null)
					ns.add(new String(e.id));
			}
			return ns;
		}
	}

	public int getNeighborsCount() {
		synchronized(cacheMutex) {
			int count = 0;
			for (int i = 0; i < cachesize; i++) {
				if (cache[i] != null)
					count++;
			}
			return count;
		}
	}

	public int getNeighborNeighborsSize(String n) {
		return -1;
	}

	public boolean isNeighbor(String source) {
		return getNeighbors().contains(source);
	}

	public void cleanLeave() {
		return;
	}

}
