/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package node;

import java.util.concurrent.locks.ReentrantLock;

import jsc.distributions.Distribution;
import jsc.distributions.Normal;
import jsc.distributions.Uniform;
import pheromone.Pheromone;
import services.IInformationSystemService;
import services.ILoadManagementService;
import services.IResourceManagementService;
import services.ISmartGridProxy;
import services.InformationSystemService;
import services.LoadManagementService;
import services.ResourceManagementService;
import services.SmartGridProxy;
import services.dss.DynamicSchedulingService;
import services.dss.profile.ProfileGenerator;
import services.dss.profile.ResourceProfile;
import support.INetworkNode;
import support.Parameters;
import support.Stats;

public abstract class Node implements INetworkNode {

	protected final Pheromone gamma;
	protected final Pheromone beta;
	protected final String id;
	private long time = 0;
	private final ReentrantLock lock = new ReentrantLock();
	private boolean shouldVisit = true;
	private final ILoadManagementService lms;
	private final IResourceManagementService rms;
	private final IInformationSystemService iss;
	private final DynamicSchedulingService dss;
	//private final ILocalStatsService ilss;
	
	private ResourceProfile localProfile;
	
	private static Distribution normal = null;
	
	private static ProfileGenerator p = null;
	
	private Distribution getNormalDistribution() {
		if (normal == null) {
			double av = (Parameters.DSS_MAX_JOBTIME_MS - Parameters.DSS_MIN_JOBTIME_MS) / 2;
			double stdev = av * 0.5;
			av = Parameters.DSS_MIN_JOBTIME_MS + av;
			normal = new Normal(av, stdev);
		}
		return normal;
	}
	
	private ProfileGenerator getProfileGenerator() {
		if (p == null) {
			p = new ProfileGenerator(false, new Uniform(), getNormalDistribution());
		}
		return p;
	}
	
	private final ISmartGridProxy sgp;

	public Node(String identifier) {
		id = identifier;
		
		ProfileGenerator pgen = getProfileGenerator();
		localProfile = pgen.generate(getIdentifier());
		
		beta = new Pheromone("beta", id);
		gamma = new Pheromone("gamma", id);
		lms = new LoadManagementService(this);
		rms = new ResourceManagementService(this);
		
		if (Parameters.DSS)
			dss = new DynamicSchedulingService(this);
		else
			dss = null;
		
		if (Parameters.ISS_ENABLED)
			iss = new InformationSystemService(this, Parameters.ISS_CACHE_SIZE);
		else 
			iss = null;
		
		if (Parameters.NOTIFY)
			System.out.format("Created %s\n", id);
		Stats.newNode();
		
		//ilss = new LocalStatsService(this);
		
		if (Parameters.SMARTGRID_PROXY) {
			sgp = new SmartGridProxy(this);
		} else {
			sgp = null;
		}
	}
	
	public ResourceProfile getProfile() {
		return localProfile;
	}
	
	public final ISmartGridProxy getSmartGridProxy() {
		return sgp;
	}

	public final ILoadManagementService getLoadManagementService() {
		return lms;
	}

	public final IResourceManagementService getResourceManagementService() {
		return rms;
	}
	
	public final IInformationSystemService getInformationSystemService() {
		return iss;
	}
	
	public final DynamicSchedulingService getDSSService() {
		return dss;
	}
	
	public final void acquire() {
		lock.lock();
	}

	public final void release() {
		lock.unlock();
	}
	
	private int step = 0;
	private int istep = 0;
	
	public void step() {
		
		time += Parameters.MS_PER_ITER;
		
		if (istep % Parameters.TRAIL_DECAY_STEP == 0) {
			iterstep();
			istep = 0;
		}
		
		istep++;
		
		if (step % Parameters.NODE_RESOLUTION_INTERVAL_ITER == 0) {
			
			timestep();
			
			if (rms != null)
				rms.cleanup();
			
			if (iss != null)
				iss.cleanup();
			
			if (Parameters.ISS_ENABLED || Parameters.ADAPTIVE_ISS)
				iss.timestep();
			
			if (Parameters.PROFILE_CACHE)
				rms.timestep();
			
			if (Parameters.DSS)
				dss.timestep();
		
			step = 0;
		}
		
		step++;
	}
	
	public abstract void timestep();
	
	public abstract void iterstep();

	public final long getTime() {
		return new Long(time) + 0l;
	}
	
	protected final void setTime(long value) {
		time = value;
	}

	public final Pheromone getGamma() {
		return gamma;
	}

	public final Pheromone getBeta() {
		return beta;
	}

	public final String getIdentifier() {
		return id;
	}

	public final boolean shouldVisit() {
		if (shouldVisit) {
			shouldVisit = false;
			return true;
		}
		return false;
	}

	public final void clearVisit() {
		shouldVisit = true;
	}

}
