/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;

import messages.SmartGridQuery;
import node.Node;
import support.Parameters;
import support.SmartGridSupport;
import ants.AntSystem;
import ants.SmartGridQueryAnt;
import bebop.client.Client;
import bebop.lib.IMessage;
import bebop.lib.Message;
import bebop.mailbox.MailBox;

public class SmartGridProxy implements ISmartGridProxy, Observer {
	
	private Map<String, Object> profile = null;
	private MailBox mailbox;
	private Node node;
	private List<String> queries;
	
	public SmartGridProxy(Node n) {
		queries = new LinkedList<String>();
		this.node = n;
		this.mailbox = SmartGridSupport.getMailService().createMailBox(this, n.getIdentifier());
	}

	public Map<String, Object> getProfile() {
		return profile;
	}

	public boolean match(Map<String, Object> query) {
		if (profile == null) return false;
		for (String k : query.keySet()) {
			if (k.equals("qid")) continue;
			if (k.equals("requester")) continue;
			if (profile.containsKey(k)) {
				if (profile.get(k) instanceof Integer) {
					Integer value = (Integer) profile.get(k);
					Integer request = (Integer) query.get(k);
					if (value < request) {
						//System.out.println("Query asked "+k+" for "+request+" got "+value);
						return false;
					}
				} else {
					//System.out.println("Query asked "+k+" for "+query.get(k)+" got "+ profile.get(k)+" match?"+profile.get(k).equals(query.get(k)));
					boolean nomatch = !profile.get(k).equals(query.get(k));
					if (nomatch) return false;
				}
			} else {
				System.out.println("Key "+k+" not matched");
				return false;
			}
		}
		//System.out.println("Node "+node.getIdentifier()+" matches query "+query);
		sendReply((String)query.get("requester"), (String)query.get("qid"));
		return true;
	}
	
	private void sendReply(String address, String qid) {
		//System.out.println("Sending reply to "+address);
		Map<String,Object> m = new HashMap<String,Object>();
		m.put("command", "queryreply");
		m.put("qid",qid);
		m.put("match", node.getIdentifier());
		Message msg = new Message("global", m);
		try {
			Client.sendMessage(address, msg);
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public boolean notify(Map<String, Object> message) {
		String recipient = "dwi";
		String address = String.format("bebop://%s:%d", SmartGridSupport.DWIHOST, SmartGridSupport.DWIPORT);
		message.put("command", "ping");
		message.put("source", node.getIdentifier());
		Message msg = new Message(recipient, message);
		try {
			Client.sendMessage(address, msg);
			return true;
		} catch (Exception e) {
			//e.printStackTrace();
			return false;
		}
	}
	
	public boolean notifyQuery(String id) {
		if (!queries.contains(id)) {
			queries.add(id);
			if (queries.size() > 1024) {
				queries.remove(0);
			}
			return true;
		}
		return false;
	}
	

	public synchronized void update(Observable source, Object data) {
		try {
			if (mailbox == null) return;
			if (source == mailbox) {
				for (;;) {
					IMessage msg = mailbox.popMessage();
					if (msg != null) {
						String command = (String) msg.getContents().get("command");
						if (command.equals("profile")) {
							//System.out.format("SmartGrid profile update for node %s: %s\n",node.getIdentifier(), msg.getContents().toString());
							Map<String,Object> np = msg.getContents();
							np.remove("command");
							np.remove("source");
							profile = np;
						} else if (command.equals("query")) {
							//System.out.format("SmartGrid query from node %s: %s\n",node.getIdentifier(), msg.getContents().toString());
							Map<String,Object> np = msg.getContents();
							np.remove("command");
							np.remove("source");
							startQuery(np);
						} else {
							//System.out.println("Received unknown message on node "+node.getIdentifier()+" with label "+label);
						}
					} else break;
				}
			}
		} catch (Exception e) {
			
		}
	}
	
	private void startQuery(Map<String, Object> query) {
		//System.out.println("Starting query on node "+node.getIdentifier()+": "+query);
		SmartGridQuery sgq = new SmartGridQuery(node.getIdentifier(), query);
		AntSystem.schedule(new SmartGridQueryAnt(new String(node.getIdentifier()), null, new String(node.getIdentifier()),
				sgq, Parameters.SEARCH_LIMIT_TTL));
	}

}
