/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package services.dss;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import jsc.distributions.Distribution;
import jsc.distributions.Normal;
import services.dss.JobStatusManager.Status;
import services.dss.monitor.DSSMonitorStats;
import services.dss.profile.ResourceProfile;
import services.dss.scheduler.ARScheduler;
import services.dss.scheduler.EDFScheduler;
import services.dss.scheduler.FIFOScheduler;
import services.dss.scheduler.FairARPoolScheduler;
import services.dss.scheduler.FairARScheduler;
import services.dss.scheduler.IScheduler;
import services.dss.scheduler.SJFScheduler;
import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;


/**
 * To solve fault tolerance:
 * 	- assignee regularly ping the initiator:
 * 		if the initiator is dead, kill the jobs
 *  - initiators regularly ping the last known assignees:
 *  	if one of the assignees are dead, revoke the job / pool
 *  If transient errors of the initiator exist, the job might be killed
 *  on some node and kept running on others but as soon as the initiator successfully
 *  detects one of the assignee fault, it will revoke all jobs.
 *  
 *  Pings from assignees to initiators can be in the form of STATUS messages,
 *  re-informing about the SCHEDULED status.
 *  
 *  If the initiator receives a STATUS message for an unknown job, it will
 *  reply with a REVOKE.
 *  
 *  Reliable communication between nodes is necessary!
 *  
 *  Ordered FIFO communication can be achieved easily with timestamps:
 *  
 *  STATUS messages from different ASSIGNEES can be ordered:
 *  	- the timestamp is: assign_count.timestamp,
 *  		for example: 3.56235435 means the 3rd assign and 
 *  		56235435 is the timestamp
 *  STATUS messages from the initiator are ordered by just using the
 *  timestamp of the initiator (as it will not change)
 *  
 *  Nodes can keep track of the status timestamp and ignore older messages
 *  (this avoids receiving multiple SCHEDULED messages in the wrong order and
 *  loosing track of the real location of the job)
 *  
 *  If one job of the pool is lost, it can be recreated (with a different UUID, that
 *  will replace the other one) and re-submitted. If the old job is still alive it will
 *  be revoked as soon as a periodic status update is sent to the initiator.
 *  
 * @author attila
 *
 */

public class DynamicSchedulingService {

	private AcceptOfferManager aom;

	private INetworkNode d;

	private IScheduler s;
	
	private IStatusManager jsm;
	
	private List<String> queries;

	public DynamicSchedulingService(INetworkNode d) {
		this.d = d;
		if (Parameters.DSS_AR_FAIR_POOL) {
			s = new FairARPoolScheduler(d, d.getProfile().cpuspeed);
			jsm = new PoolStatusManager(d, new JobStatusManager(s));
		} else if (Parameters.DSS_AR_FAIR) {
			s = new FairARScheduler(d, d.getProfile().cpuspeed);
		} else if (Parameters.DSS_AR) {
			s = new ARScheduler(d, d.getProfile().cpuspeed);
		} else if (Parameters.DSS_EDF) {
			s = new EDFScheduler(d, d.getProfile().cpuspeed);
		} else if (Randomizer.pop() < Parameters.FIFO_PROBABILITY) {
			s = new FIFOScheduler(d, d.getProfile().cpuspeed);
		} else {
			s = new SJFScheduler(d, d.getProfile().cpuspeed);
		}
		
		if (jsm == null)
			jsm = new BogusJobStatusManager();
		
		aom = new AcceptOfferManager(d, Parameters.ACCEPT_OFFER_WAIT_MS);
		queries = new LinkedList<String>();
		
		
		System.out.println("[RESOURCE] Node "+d.getIdentifier()+" got scheduler "+ s.getSchedulerName() +" assigned profile "+d.getProfile());
		DSSMonitorStats.init();
		DSSMonitorStats.addDaemon(d, this);
	}
	
	public IStatusManager getJobStatusManager() {
		return jsm;
	}

	public IScheduler getScheduler() {
		return s;
	}
	
	public AcceptOfferManager getAcceptOfferManager() {
		return aom;
	}

	public synchronized void accept(String jobid, String rnode, Number remoteCost) {
		if (aom.isWaitingJob(jobid)) {
			aom.pushOffer(jobid, rnode, remoteCost.longValue());
		} else if (s.isScheduledJob(jobid)) {
			Long localCost = s.getScore(s.getProfile(jobid));
			if (remoteCost.longValue() < localCost) {
				ResourceProfile prof = s.getProfile(jobid);
				if (s.isReservationScheduling() && (prof.tempDelay <= 0)) return; // Don't reassign if the job is not delayed
				ResourceProfile rp = s.unschedule(jobid);
				if (rp != null) { // Null if the job is already
					rp.promisedCosts.add(0, remoteCost.longValue());
					rp.delays.add(0,rp.tempDelay);
					DSSStats.reassigned(jobid, rnode);
					rp.assigns += 1;
					Messaging.sendAssign(d, rp, rnode);
				}
			}
		}
	}

	public void addJob(ResourceProfile profile) {
		profile.initiator = d.getIdentifier();
		aom.startAcceptOffersFor(profile);
		Messaging.sendRequest(d, profile);
	}


	public synchronized boolean inform(String rnode, Map<String, Object> profile, Number eta) {
		if (rnode.equals(d.getIdentifier())) return false;
		ResourceProfile rp = new ResourceProfile(profile);
		if (!d.getProfile().matches(rp)) return false;
		Long localEstimation = s.getScore(rp);
		Long threshold = ((long) Math.signum(localEstimation)) * Parameters.INFORM_ACCEPT_THRESHOLD_MS;
		if (threshold == 0) {
			threshold = -1 * Parameters.INFORM_ACCEPT_THRESHOLD_MS;
		}
		if (localEstimation < (eta.longValue() - threshold)) {
			Messaging.sendAccept(d, rp.id, localEstimation, rnode);
			return true;
		}
		return false;
	}

	public boolean matches(ResourceProfile profile) {
		return d.getProfile().matches(profile);
	}
	
	public boolean request(ResourceProfile profile, String rnode) {
		if (d.getProfile().matches(profile)) {
			Long eta = s.getScore(profile);
			Messaging.sendAccept(d, profile.id, eta, rnode);
			return true;
		}
		return false;
	}

	public void assign(Map<String, Object> profile) {
		ResourceProfile jp = new ResourceProfile(profile);
		if (!s.isScheduledJob(jp.id)) {
			long score = s.getScore(jp);
			jp.actualCosts.add(0,score);
			jp.currentNode = d;
			s.schedule(jp);
			double ts = Double.parseDouble(String.format("%d.%d", jp.assigns, d.getTime()));
			Messaging.sendStatus(d, jp.id, jp.initiator, Status.SCHEDULED, ts);
		} else {
			System.out.println("Warning! Trying to reschedule already scheduled job "+jp.id+" on node "+d.getIdentifier());
		}
	}
	
	public void status(String uuid, String node, Status status, Double timestamp) {
		if (jsm.updateTimestamp(uuid, timestamp) && (status != Status.REVOKED) && (status != Status.FINISHED)) return;
		jsm.setJobStatus(uuid, status, node);
	}

	public void setScheduler(String scheduler) {
		if (scheduler.equals("sjf"))
			s = new SJFScheduler(d, d.getProfile().cpuspeed);
		else if (scheduler.equals("fifo"))
			s = new FIFOScheduler(d, d.getProfile().cpuspeed);
	}

	public void unschedule(String profileid ) {
		s.unschedule(profileid);
	}
	

	public void forceUnschedule(String profileid) {
		s.forceUnschedule(profileid);
	}
	
	/**
	 java -cp .:../../SolenopsisII/c.jar -Xms1500m -Xmx1500m algo.BlatantSAlgorithm --initial ~/topo-500 --underlying ~/lan-1281-rtt --prefix /t/tmp/blatsarls --maxiter 15000000 --fileinterval -1 --screeninterval 250 --cpufactor 1 --cinterval 250 --cretry 1000 --birthprob 0.01 --omegac 1 --omegad 1 --d 5 --rttenabled --m 8 --mo 4 --gammadecay 0.99999 --rttfactor 10 --searchstart 25000 --searchint 3000 --searchstop 100000 --limitttl 9 --limitffw 4
	 */
	
	private long lastInform = 0;
	
	public synchronized void timestep() {
		s.progress();
		aom.timestep();
		if (s.getScheduleSize() == 0)
			lastInform = NetworkSystem.getGlobalTimeMs();
		if ((NetworkSystem.getGlobalTimeMs() - lastInform) >= Parameters.INFORM_INTERVAL_MS) {
			if (Parameters.INFORM_LAST_N > 0) {
				List<String> jobs = s.getJobsForInform();
				Collections.reverse(jobs);
				int informed = 0;
				long score = 0;
				for (String job : jobs) {
					ResourceProfile r = s.getProfile(job);
					if (r.isLocked())
						continue;
					if (s.isReservationScheduling() && (r.tempDelay <= 0)) // Don't inform if the job is not delayed
						continue;
					score = s.getScore(r);
					if (score <= 0) break;
					r.informs++;
					DSSStats.inform(r.id, d.getIdentifier());
					Messaging.sendInform(d, r, score);
					informed++;
					if (informed >= Parameters.INFORM_LAST_N)
						break;
				}
				lastInform = NetworkSystem.getGlobalTimeMs();
			}
		}
	}
	
	public boolean notifyQuery(String id) {
		if (!queries.contains(id)) {
			queries.add(id);
			if (queries.size() > Parameters.DSS_QUERY_TRACK_SIZE)
				queries.remove(0);
			return true;
		}
		return false;
	}
	
	public static void main(String[] args) {
		long MAXJOB = 14400000;
		long MINJOB = 3600000;
		double av = (MAXJOB - MINJOB) / 2;
		double stdev = av * 0.5;
		av = MINJOB + av;
		Distribution d = new Normal(av, stdev);
		
		for (int i=0; i < 100; i++) {
			System.out.println(DSSStats.formatTime((long) d.random()));
		}
		
	}

}
