/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services.dss;

import java.util.HashMap;
import java.util.Map;

import services.dss.JobStatusManager.Status;
import services.dss.scheduler.pool.Pool;
import support.INetworkNode;
import support.NetworkSystem;

public class PoolStatusManager implements IStatusManager {
	
	private Map<String,Pool> pools;
	private INetworkNode d;
	private IStatusManager jsm;
	
	public PoolStatusManager(INetworkNode d, IStatusManager jsm) {
		pools = new HashMap<String,Pool>();
		this.d = d;
		this.jsm = jsm;
	}
	
	public void addPool(Pool p) {
		synchronized(pools) {
			pools.put(p.getUUID(), p);
			for (String j : p.getJobs()) {
				jsm.addJob(j);
			}
		}
	}

	public void onJobReady(String uuid, String node, Double cpuspeed) {
		synchronized(pools) {
			Pool p = getPoolContaining(uuid);
			if (p == null) {
				Messaging.sendStatus(d, uuid, node, Status.REVOKED, Double.MAX_VALUE);
			} else {
				p.onJobReady(uuid, node, cpuspeed);
			}
		}
	}
	
	private Pool getPoolContaining(String uuid) {
		for (Pool p : pools.values()) {
			if (p.hasJob(uuid)) return p;
		}
		return null;
	}

	public String getJobLocation(String uuid) {
		return jsm.getJobLocation(uuid);
	}

	public Status getJobStatus(String uuid) {
		return jsm.getJobStatus(uuid);
	}

	public void removeJob(String uuid) {
		jsm.removeJob(uuid);
	}

	public void setJobStatus(String uuid, Status s, String node) {
		jsm.setJobStatus(uuid, s, node);
		if (s == Status.REVOKED) {
			DSSStats.failed();
			removeJob(uuid);
		} else if (s == Status.FINISHED) {
			Pool p = getPoolContaining(uuid);
			if (p != null) {
				DSSStats.finishPool(p.getUUID(), p.getSize());
				pools.remove(p.getUUID());
			}
			return;
		} else if (s == Status.READY) {
			/* For simulation purposes only: 
			 * Retrieve speed of the target node, in order to
			 * let the simulator know how long to simulate execution */
			INetworkNode n = NetworkSystem.getNode(node);
			n.acquire();
			double cpuspeed= n.getProfile().cpuspeed;
			n.release();
			/* End of simulation trick */
			onJobReady(uuid, node, cpuspeed);
		} else if (s == Status.RUNNABLE) {
			d.getDSSService().getScheduler().getProfile(uuid).isRunnable = true;
		} else if (s == Status.FAILED) {
			Pool p = getPoolContaining(uuid);
			if (p != null) {
				p.revokeAll();
			}
		}
	}

	public void addJob(String uuid) {
		jsm.addJob(uuid);
	}

	public boolean updateTimestamp(String uuid, Double ts) {
		return jsm.updateTimestamp(uuid, ts);
	}

}
