/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services.dss.monitor;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

import services.dss.DSSStats;
import services.dss.DynamicSchedulingService;
import services.dss.profile.ResourceProfile;
import services.dss.scheduler.IScheduler;
import support.NetworkSystem;

public class SchedulerMonitorWidget extends JComponent implements MouseListener, MouseMotionListener, MouseWheelListener {
	
	private static final long serialVersionUID = 8255031292526847615L;

	private DynamicSchedulingService dss;
	
	private int max = 24;
	
	private int mousePosition = -1;
	
	private ResourceProfile highlightedJob = null;
	
	private Font datafont = null;
	
	private final Color[] colors = {new Color(155,155,155), Color.gray};
	
	private final Color[] delaycolors = {new Color(255,155,155), new Color(255,200,200)};

	
	private double cpuspeed;
	
	private static AlphaComposite alpha = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f);
	private static AlphaComposite full = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f);
	
	private boolean normalSpeed = false;
	
	private static boolean showRealTime = false;
	
	public SchedulerMonitorWidget(DynamicSchedulingService s) {
		this.dss = s;
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
		cpuspeed = dss.getScheduler().getCPUSpeed();
	}
	
	public void setFixedHoursShown(int max) {
		this.max = max;
	}
	
	
	public void paintComponent(Graphics g) {
		try {
			Graphics2D g2d = (Graphics2D) g;
			
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);
			g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
					RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
	
			int width = getWidth() - 1;
			double rwidth = getWidth() - 1;
			int height = getHeight() - 1;
			int gheight = height - 15;
			
			int offset = 0;
			int i = 0;
			
			IScheduler s = dss.getScheduler();
			
			double scheduleLength = 24;
			
			if (datafont == null) {
				datafont = getFont().deriveFont((float) 8);
			}
			
			g2d.setFont(datafont);
			
			if (max > 0)
				scheduleLength = max;
			else
				scheduleLength = (s.getScheduleLength() / 3600000F) + 1;
				
			boolean useAR = s.isReservationScheduling();
			
			boolean isPoolScheduler = s.getSchedulerName().equals("ARFairPool");
			
			double factor = 1.0;
			if (!normalSpeed && !useAR)
				factor = cpuspeed;
			
			g2d.setColor(Color.white);
			g2d.fillRect(0, 0, width, gheight);
			
			
			
			long cTime = NetworkSystem.getGlobalTimeMs();
			
			setHighlightedJob(null);
			for (String job : s.getJobs()) {
				ResourceProfile profile = s.getProfile(job);
				if (useAR) {
					if (!profile.isRunning())
						offset= (int) ((((profile.reservationStart + profile.tempDelay) - cTime) / 3600000F) * rwidth / scheduleLength);
					else
						offset = 0;
				}
				double base = 0;
				if (profile.isRunning()) {
					if (useAR) {
						base = ((profile.reservationEnd + profile.startDelay) - cTime) / 3600000F;
					} else {
						base = profile.remainingTime / 3600000F;
					}
				} else
					base = profile.getEstimatedRunningTime(factor) / 3600000F;
				int blockwidth = (int) (base * rwidth / scheduleLength);
				if ((offset <= mousePosition) && (mousePosition <= (offset+blockwidth))) {
					g2d.setColor(Color.orange);
					setHighlightedJob(profile);
					setToolTipText(formatTooltip(s, profile));
				} else if (profile.id.equals(DSSMonitorStats.getHighlightedJobId())) {
					g2d.setColor(Color.blue);
				} else if (profile.pool.equals(DSSMonitorStats.getHighlightedPoolId())) {
					g2d.setColor(Color.cyan);
				} else if (profile.isRunning()) {
					g2d.setColor(Color.green);
				} else if (isPoolScheduler && profile.isReady() && !profile.isRunning()) {
					g2d.setColor(Color.magenta);
				} else {
					if (profile.tempDelay <= 0)
						g2d.setColor(colors[i%2]);
					else
						g2d.setColor(delaycolors[i%2]);
				}
				g2d.setComposite(alpha); 
				g2d.fillRect(offset, 0, blockwidth, gheight);
				int middlePoint = offset + blockwidth / 2 - 5;
				g2d.setComposite(full);
				g2d.setColor(Color.white);
				g2d.drawString(String.format("%d", profile.assigns), middlePoint, gheight / 2 +4);
				
				if (profile.tempDelay > 0)
					g2d.setColor(Color.red);
				else
					g2d.setColor(Color.black);
				g2d.drawRect(offset, 0, blockwidth, gheight);
				offset += blockwidth;
				i++;
			}
			
			g2d.setColor(getForeground());
			int tick = 0;
			
			for (int h=0; h<scheduleLength; h += 2) {
				tick = (int) (h * width / scheduleLength);
				g2d.drawLine(tick, gheight, tick, height -5);
				if (showRealTime) {
					g2d.drawString(String.format("@%d%s", h + (int) cTime / 3600000, normalSpeed ? "N" : ""), tick+2, height -3);
				} else {
					g2d.drawString(String.format("%d%s", h, normalSpeed ? "N" : ""), tick+2, height -3);
				}
				
			}
			
			g2d.setColor(Color.black);
			g2d.drawRect(0, 0, width, gheight);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void setHighlightedJob(ResourceProfile profile) {
		highlightedJob = profile;
	}

	public static String formatTooltip(IScheduler s, ResourceProfile profile) {
		String optional = "";
		Long completion = s.getEstimatedCompletionTime(profile);
		Long waiting = completion - profile.getEstimatedRunningTime(s.getCPUSpeed());
		if (profile.isRunning())
			optional = String.format("<tr><td><b>Started at</b>:</td><td>%s</td></tr>" +
					"<tr><td><b>Remaining</b>:</td><td>%s</td></tr>" +
					"<tr><td><b>Remaining</b>:</td><td>%s</td></tr>", DSSStats.formatTime(profile.startedAt), DSSStats.formatTime(profile.remainingTime), DSSStats.formatTime(profile.startDelay));
		else
			optional = String.format("<tr><td><b>Time to completion</b>:</td><td>%s</td></tr>" +
					"<tr><td><b>Time to start</b>:</td><td>%s</td></tr>", DSSStats.formatTime(completion), DSSStats.formatTime(waiting));
		return String.format("<html><table border=\"0\">" +
				"<tr><td><b>Job id</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Pool</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Pool Size</b>:</td><td>%d</td></tr>" +
				"<tr><td><b>Node</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Created</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Arch</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>OS</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Version</b>:</td><td>%d</td></tr>" +
				"<tr><td><b>Memory</b>:</td><td>%d</td></tr>" +
				"<tr><td><b>Disk</b>:</td><td>%d</td></tr>" +
				"<tr><td><b>ERT</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>ERT(1)</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Drift</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Assigns</b>:</td><td>%d</td></tr>" +
				"<tr><td><b>Informs</b>:</td><td>%d</td></tr>" +
				"<tr><td><b>Deadline</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Reservation Start</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Reservation End</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>ACosts</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>PCosts</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Balance</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Delays</b>:</td><td>%s</td></tr>" +
				"<tr><td><b>Actual delay</b>:</td><td>%s</td></tr>" +
				"%s" +
				"</table></html>", 
				profile.id, profile.pool, profile.poolsize, profile.currentNode == null ? "n/a" : profile.currentNode.getIdentifier(), DSSStats.formatTime(profile.createdAt), profile.arch, profile.os, profile.minimumVersion, 
				profile.memory, profile.disk, DSSStats.formatTime(profile.getEstimatedRunningTime(s.getCPUSpeed())),
				DSSStats.formatTime(profile.getEstimatedRunningTime(1.0)),
				DSSStats.formatTime(profile.drift),
				profile.assigns, profile.informs, DSSStats.formatTime(profile.deadline), DSSStats.formatTime(profile.reservationStart), 
				DSSStats.formatTime(profile.reservationEnd), 
				profile.actualCosts.toString(),
				profile.promisedCosts.toString(),
				getCostBalanceString(profile.promisedCosts, profile.actualCosts),
				profile.delays.toString(),
				DSSStats.formatTime(profile.tempDelay),
				optional);
	}
	
	public static String getCostBalanceString(List<Long> promised, List<Long> actual) {
		StringBuilder sb = new StringBuilder("Last: ");
		for (int i = 0; i < promised.size(); i++) {
			sb.append(actual.get(i) - promised.get(i));
			sb.append(" ");
		}
		return sb.toString();
	}

	public void mouseClicked(MouseEvent arg0) {
		if (arg0.getClickCount() == 2) {
			if (arg0.isControlDown()) {
				max = -1;
				repaint();
			} else if (arg0.isShiftDown()) {
				showRealTime = !showRealTime;
			} else {
				normalSpeed = !normalSpeed;
				repaint();
			}
		}
	}

	public void mouseEntered(MouseEvent arg0) {
	}

	public void mouseExited(MouseEvent arg0) {
		mousePosition = -1;
		setToolTipText(null);
		setHighlightedJob(null);
		repaint();
	}

	public void mousePressed(MouseEvent arg0) {
		if (arg0.getButton() == 3) {
			JPopupMenu popup = new JPopupMenu();
			
			if (DSSMonitorStats.getHighlightedJob() != null) {
				JMenuItem clearhighlight = new JMenuItem("Clear highlight", new Integer(KeyEvent.VK_C));
				clearhighlight.addActionListener(new AbstractAction() {

					private static final long serialVersionUID = -8216395395607481262L;

					public void actionPerformed(ActionEvent e) {
						DSSMonitorStats.setHighlightedJob(null);
					}

				});
				popup.add(clearhighlight);
			}
			
			if (DSSMonitorStats.getHighlightedPoolId() != null) {
				JMenuItem clearhighlight = new JMenuItem("Clear pool highlight", new Integer(KeyEvent.VK_C));
				clearhighlight.addActionListener(new AbstractAction() {

					private static final long serialVersionUID = -8216395395607481262L;

					public void actionPerformed(ActionEvent e) {
						DSSMonitorStats.setHighlightedPoolId(null);
					}

				});
				popup.add(clearhighlight);
			} else {
				
				final ResourceProfile job = highlightedJob;

				JMenuItem highlight = new JMenuItem("Highlight pool", new Integer(KeyEvent.VK_H));
				highlight.addActionListener(new AbstractAction() {

					private static final long serialVersionUID = 7812745379992574091L;

					public void actionPerformed(ActionEvent e) {
						DSSMonitorStats.setHighlightedPoolId(job.pool);
					}

				});
				popup.add(highlight);
			}

			
			if (highlightedJob != null) {
				
				final ResourceProfile job = highlightedJob;

				JMenuItem highlight = new JMenuItem("Highlight", new Integer(KeyEvent.VK_H));
				highlight.addActionListener(new AbstractAction() {

					private static final long serialVersionUID = 7812745379992574091L;

					public void actionPerformed(ActionEvent e) {
						DSSMonitorStats.setHighlightedJob(job);
					}

				});
				popup.add(highlight);

				JMenuItem unschedule = new JMenuItem("Unschedule", new Integer(KeyEvent.VK_U));
				unschedule.addActionListener(new AbstractAction() {

					private static final long serialVersionUID = -8216395395607481262L;

					public void actionPerformed(ActionEvent e) {
						dss.getScheduler().forceUnschedule(job.id);
					}

				});
				popup.add(unschedule);
				
				
				JMenuItem printInfo = new JMenuItem("Print info", new Integer(KeyEvent.VK_U));
				printInfo.addActionListener(new AbstractAction() {

					private static final long serialVersionUID = -8216395395607481262L;

					public void actionPerformed(ActionEvent e) {
						System.out.println(formatTooltip(dss.getScheduler(), job));
					}
				});
				popup.add(printInfo);
				
			}
			
			JMenuItem showDetail = new JMenuItem("Show detail", new Integer(KeyEvent.VK_U));
			showDetail.addActionListener(new AbstractAction() {

				private static final long serialVersionUID = -8216395395607481262L;

				public void actionPerformed(ActionEvent e) {
					DSSMonitorStats.toggleDetailWindow();
				}
			});
			popup.add(showDetail);
			
			
			
			
			popup.show(this, arg0.getX(), arg0.getY());	
		}
	}

	public void mouseReleased(MouseEvent arg0) {
	}

	public void mouseDragged(MouseEvent e) {
	}

	public void mouseMoved(MouseEvent e) {
		mousePosition = e.getX();
		repaint();
	}

	public void mouseWheelMoved(MouseWheelEvent arg0) {
		max = Math.max(0,max + arg0.getWheelRotation());
		repaint();
	}


}
