/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package services.dss.scheduler;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import services.dss.DSSStats;
import services.dss.Messaging;
import services.dss.JobStatusManager.Status;
import services.dss.profile.ResourceProfile;
import support.INetworkNode;
import support.Parameters;
import support.Randomizer;

public class EDFScheduler implements IScheduler {

	long currentTime = 0l;

	List<ResourceProfile> schedule = new ArrayList<ResourceProfile>();
	
	private double cpuspeed = 1;
	private INetworkNode d;
	
	private static Comparator<ResourceProfile> deadlineComparator = new Comparator<ResourceProfile>() {

		public int compare(ResourceProfile arg0, ResourceProfile arg1) {
			if (arg0.isRunning()) return -1;
			return arg0.deadline.compareTo(arg1.deadline);
		}
		
	};
	
	private static Comparator<ResourceProfile> slacktimeComparator = new Comparator<ResourceProfile>() {

		public int compare(ResourceProfile arg0, ResourceProfile arg1) {
			return arg1.tempSlack.compareTo(arg0.tempSlack);
		}
		
	};
	
	public EDFScheduler(INetworkNode d, double cpuspeed) {
		this.cpuspeed  = cpuspeed;
		this.d = d;
	}
	
	public double getCPUSpeed() {
		return cpuspeed;
	}
	
	private Long getETC(List<ResourceProfile> temp, ResourceProfile j) {
		long waitingTime = 0;
		for (int i = 0; i < temp.size(); i++) {
			ResourceProfile f = temp.get(i);
			if (!f.isRunning())
				waitingTime += f.getEstimatedRunningTime(cpuspeed);
			else
				waitingTime += Math.max(0,f.getEstimatedRunningTime(cpuspeed)
						- (currentTime - f.startedAt));
			if (f == j) break;
		}
		return waitingTime;
	}

	public Long getEstimatedCompletionTime(ResourceProfile j) {
		List<ResourceProfile> temp = new ArrayList<ResourceProfile>();
		synchronized(schedule) {
			temp.addAll(schedule);
			if (!temp.contains(j))
				temp.add(j);
			Collections.sort(temp, deadlineComparator);
		}
		return getETC(temp, j);
	}

	public List<String> getJobs() {
		List<String> jid = new ArrayList<String>();
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				jid.add(f.id);
			}
		}
		return jid;
	}

	public ResourceProfile getProfile(String jobid) {
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.id.equals(jobid))
					return f;
			}
		}
		return null;
	}

	public boolean isScheduledJob(String jobid) {
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.id.equals(jobid))
					return true;
			}
		}
		return false;
	}

	public synchronized void progress() {
		currentTime = d.getTime();
		synchronized (schedule) {
			if (schedule.size() == 0)
				return;
			ResourceProfile running = schedule.get(0);
			if (!running.isRunning()) {
				running.startedAt = currentTime;
				running.setRunning();
				/* Notify status to initiator */
				double ts = Double.parseDouble(String.format("%d.%d", running.assigns, d.getTime()));
				Messaging.sendStatus(d, running.id, running.initiator, Status.RUNNING, ts);
				/* End notify to initiator */
				if (running.actualRunningTime != -1) {
					running.remainingTime = (long) (running.actualRunningTime / cpuspeed);
					running.drift = 0l;
				} else {
					running.remainingTime = running.getEstimatedRunningTime(cpuspeed);
					running.drift = (long) ((Randomizer.pop() - 0.5) * running.getEstimatedRunningTime(cpuspeed) * (Parameters.DSS_DRIFT * 2)); // Plus minus 10%
					if (Parameters.BAD_DRIFT_ONLY)
						running.drift = Math.abs(running.drift);
				}
				DSSStats.started(running.id);
			} else {
				running.remainingTime = running.getEstimatedRunningTime(cpuspeed) - (currentTime - running.startedAt);
				if ((running.remainingTime + running.drift) <= 0l) {
					running.remainingTime = 0l;
					running.finishedAt = currentTime;
					schedule.remove(0);
					DSSStats.finished(running, cpuspeed);
					DSSStats.deadline(currentTime, running.deadline);
					
					/* Notify status to initiator */
					double ts = Double.parseDouble(String.format("%d.%d", running.assigns, d.getTime()));
					Messaging.sendStatus(d, running.id, running.initiator, Status.FINISHED, ts);
					/* End notify to initiator */
				}		
			}
		}
	}

	public void schedule(ResourceProfile j) {
		synchronized (schedule) {
			schedule.add(j);
			Collections.sort(schedule, deadlineComparator);
		}
	}

	public ResourceProfile unschedule(String jobid) {
		synchronized (schedule) {
			ResourceProfile j = getProfile(jobid);
			if ((j != null) && (!j.isLocked()))
				if (schedule.remove(j)) {
					return j;
				}
		}
		return null;
	}
	
	public ResourceProfile forceUnschedule(String jobid) {
		synchronized (schedule) {
			ResourceProfile j = getProfile(jobid);
			if (j != null)
				if (schedule.remove(j)) {
					return j;
				}
		}
		return null;
	}

	public Long getScheduleLength() {
		long length = 0l;
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.isRunning())
					length += f.remainingTime;
				else
					length += f.getEstimatedRunningTime(cpuspeed);
			}
		}
		return length;
	}
	
	public int getScheduleSize() {
		synchronized (schedule) {
			return schedule.size();
		}
	}

	public String getSchedulerName() {
		return "EDF";
	}
	
	public synchronized Long getScore(ResourceProfile j) {
		List<ResourceProfile> temp = new ArrayList<ResourceProfile>();
		synchronized(schedule) {
			temp.addAll(schedule);
			if (!temp.contains(j))
				temp.add(j);
			Collections.sort(temp, deadlineComparator);
		}
		return computeScore(temp);
	}
	
	private Long computeScore(List<ResourceProfile> temp) {
		boolean schedulable = true;
		List<Long> slacktimes = new ArrayList<Long>();
		
		long rpetc = 0;
		for (ResourceProfile f : temp) {
			if (!f.isRunning())
				rpetc += f.getEstimatedRunningTime(cpuspeed);
			else
				rpetc += Math.max(0,f.getEstimatedRunningTime(cpuspeed)
						- (currentTime - f.startedAt));
			Long timeOfCompletion = currentTime + rpetc;
			Long slackTime = f.deadline - timeOfCompletion; 
			slacktimes.add(slackTime);
			if (slackTime < 0) {
				schedulable = false;
			}
		}
		

		Long score = 0l;
		for (Long st : slacktimes) {
			if ((!schedulable && (st < 0l)) || schedulable) {
				score += Math.abs(st);
			}
		}
		int sign = schedulable ? -1 : 1;
		return sign * score;
	}

	public synchronized List<String> getJobsForInform() {
		List<ResourceProfile> temp = new ArrayList<ResourceProfile>();
		ArrayList<String> jid = new ArrayList<String>();
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				temp.add(f);
			}
			long rpetc = 0;
			for (ResourceProfile f : temp) {
				if (!f.isRunning())
					rpetc += f.getEstimatedRunningTime(cpuspeed);
				else
					rpetc += Math.max(0,f.getEstimatedRunningTime(cpuspeed)
							- (currentTime - f.startedAt));
				Long timeOfCompletion = currentTime + rpetc;
				f.tempSlack = f.deadline - timeOfCompletion; 				
			}
			//System.out.println("---------------");
			Collections.sort(temp, slacktimeComparator);
			for (ResourceProfile f : temp) {
				//System.out.println(f.id+"-slack->"+f.tempSlack);
				jid.add(f.id);
				f.tempSlack = 0l;
			}
			//System.out.println("---------------");
		}
		return jid;
	}
	
	public static void main(String[] args) {
		ResourceProfile j1 = new ResourceProfile();
		j1.id = "One";
		j1.setEstimatedRunningTime(10l);
		j1.deadline = 100l;
		
		ResourceProfile j2 = new ResourceProfile();
		j2.id = "Two";
		j2.setEstimatedRunningTime(20l);
		j2.deadline = 90l;
		
		ResourceProfile j3 = new ResourceProfile();
		j3.id = "Three";
		j3.setEstimatedRunningTime(10l);
		j3.deadline = 75l;
		
		ResourceProfile j4 = new ResourceProfile();
		j4.id = "Four";
		j4.setEstimatedRunningTime(25l);
		j4.deadline = 50l;
		
		ResourceProfile j5 = new ResourceProfile();
		j5.id = "Five";
		j5.setEstimatedRunningTime(90l);
		j5.deadline = 90l;
		
		IScheduler s = new EDFScheduler(null, 1);
		//IScheduler s = new SJFScheduler(1);
		
		System.out.println(s.getScore(j1));
		s.schedule(j1);
		System.out.println("Inform:"+s.getJobsForInform());
		System.out.println(s.getJobs());
		System.out.println(s.getScore(j2));
		s.schedule(j2);
		System.out.println("Inform:"+s.getJobsForInform());
		System.out.println(s.getJobs());
		System.out.println(s.getScore(j3));
		s.schedule(j3);
		System.out.println("Inform:"+s.getJobsForInform());
		System.out.println(s.getJobs());
		System.out.println(s.getScore(j4));
		s.schedule(j4);
		System.out.println("Inform:"+s.getJobsForInform());
		System.out.println(s.getJobs());
		System.out.println(s.getScore(j5));

		s.schedule(j5);
		System.out.println("Inform:"+s.getJobsForInform());
		System.out.println(s.getJobs());
		
	}
	
	public boolean isDeadlineScheduling() {
		return false;
	}

	public boolean isReservationScheduling() {
		return false;
	}

}
