/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package syncutil;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

import logging.StaticLogger;
import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import support.Stats;
import util.Dijkstra;
import util.Graph;

public class SequentialCompute implements ICompute {

	private int diameter = 0;
	private int radius = 0;
	private CyclicBarrier b;
	private List<Integer> lengths;
	private float apl = 0;

	private class Computer implements Runnable {

		public void run() {
			for (String job : nodes) {
				Map<String, Integer> dist = distancesFromSource(job);
				int retval = Collections.max(dist.values());
				pushResult(retval);
				//Algorithm.star.progress();
				NetworkSystem.clearVisit();
			}
			try {
				b.await();
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (BrokenBarrierException e) {
				e.printStackTrace();
			}
		}
	}

	private Map<String, Integer> distancesFromSource(String source) {
		Map<String, Integer> visited = new Hashtable<String, Integer>();
		Set<String> toVisit = new HashSet<String>();
		Set<String> visiting = new HashSet<String>();
		int distance = 0;
		int nodesvisited = 0;

		toVisit.add(source);

		while (!toVisit.isEmpty()) {
			visiting.clear();
			visiting.addAll(toVisit);
			toVisit.clear();
			for (String enne : visiting) {
				String n = new String(enne);
				INetworkNode node = NetworkSystem.getNode(n);
				if ((node != null) && (node.shouldVisit())) {
					visited.put(n, new Integer(distance));
					nodesvisited++;
					node.acquire();
					for (String k : node.getNeighbors()) {
						INetworkNode u = NetworkSystem.getNode(k);
						if (u != null)
							if (!Parameters.COMPLETE_NS_LCC
									|| u.getNeighbors().contains(n))
								toVisit.add(new String(k));
					}
					node.release();
				}
			}
			distance++;
		}

		if ((nodesvisited != NetworkSystem.getNodeCount())) {
			// System.out.format("Network is partitioned! %d nodes, %d visited
			// by BFS; job from %s, neighbors %d \n",
			// NetworkSystem.getNodeCount(), nodesvisited, source,
			// NetworkSystem.getNode(source).getNeighborsCount());
			Stats.partitioned();
		} else
			Stats.unpartitioned();

		if (nodesvisited - NetworkSystem.getNodeCount() == 1) {
			System.out
					.format(
							"Network is partitioned! %d nodes, %d visited by BFS; job from %s, neighbors %d \n",
							NetworkSystem.getNodeCount(), nodesvisited, source,
							NetworkSystem.getNode(source).getNeighborsCount());
		}

		Stats.partition(nodesvisited);
		if ((nodesvisited != NetworkSystem.getNodeCount())) {
			Stats.partitioned();
			if (Parameters.DEBUG) {
				StaticLogger
						.log(
								"Network is partitioned! %d nodes, %d visited by BFS; job from %s, neighbors %d \n",
								NetworkSystem.getNodeCount(), nodesvisited,
								source, NetworkSystem.getNode(source)
										.getNeighborsCount());

				Graph g = new Graph(true);
				for (String key : NetworkSystem.getNodes()) {
					g.addVertex(new String(key));
					Collection<String> ns = NetworkSystem.getNode(key)
							.getNeighbors();
					for (String n : ns) {
						g.addEdge(new String(key), new String(n));
					}
				}

				for (String key : NetworkSystem.getNodes()) {
					if (!visited.containsKey(key)) {
						List<String> path = Dijkstra.getShortestPath(g, source,
								key);
						Collection<String> ns = NetworkSystem.getNode(key)
								.getNeighbors();
						StaticLogger.log("Starting from " + source + ", " + key
								+ " is not reachable - neighbors are " + ns);
						StaticLogger.log("Path is " + path);
						if (path.size() != 1) {
							System.out.println("Error during BFS computation!");
							System.exit(1);
						}
						for (String t : ns) {
							if (NetworkSystem.existNode(t))
								StaticLogger.log("\t\tNeighbors of "
										+ t
										+ " are "
										+ NetworkSystem.getNode(t)
												.getNeighbors());
						}
					}
				}
			}
		} else
			Stats.unpartitioned();

		return visited;
	}

	private Stack<String> nodes;
	private int results;

	public SequentialCompute() {
		nodes = new Stack<String>();
		b = new CyclicBarrier(2);
		lengths = new ArrayList<Integer>();
	}

	private synchronized void pushResult(int val) {
		lengths.add(val);
		diameter = Math.max(diameter, val);
		radius = Math.min(radius, val);
		results--;
	}

	public void computeBFS() {
		nodes.clear();
		if (Parameters.SAMPLE_COMPUTE > 0) {
			int count = Math.min(Parameters.SAMPLE_COMPUTE, NetworkSystem.getNodeCount());
			nodes.addAll(NetworkSystem.getNodesSample(count));
		} else {
			nodes.addAll(NetworkSystem.getNodes());
		}
		results = nodes.size();
		radius = Integer.MAX_VALUE;
		lengths.clear();
		diameter = 0;
		new Thread(new Computer()).start();
		try {
			b.await();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}
		computeAPL();
	}

	public int getDiameter() {
		return diameter;
	}

	public int getRadius() {
		return radius;
	}

	private void computeAPL() {
		apl = 0;
		for (int i : lengths) {
			apl = apl + (float) i;
		}
		apl = apl / (float) lengths.size();
	}

	public float getAveragePathLength() {
		return apl;
	}

}
