/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

public class AStar {

	/*
	 * A*, shortest path Adapted from the Wikipedia, The Free Encyclopedia
	 * (http://en.wikipedia.org/wiki/A*_search_algorithm)
	 */

	class ThreeDimensionalHashtable<K, V> {

		private Hashtable<K, Hashtable<K, V>> data;
		private V nullvalue;

		public ThreeDimensionalHashtable(V nullvalue) {
			data = new Hashtable<K, Hashtable<K, V>>();
			this.nullvalue = nullvalue;
		}

		public void put(K key1, K key2, V value) {
			if (data.containsKey(key1)) {
				data.get(key1).put(key2, value);
			} else {
				Hashtable<K, V> entry = new Hashtable<K, V>();
				entry.put(key2, value);
				data.put(key1, entry);
			}
		}

		public V get(K key1, K key2) {
			if (data.containsKey(key1)) {
				Hashtable<K, V> entry = data.get(key1);
				if (entry.containsKey(key2))
					return entry.get(key2);
			}
			return nullvalue;
		}

		public int size() {
			return data.size();
		}

	}

	private ThreeDimensionalHashtable<String, Integer> distances;
	private Graph graph;

	public AStar(Graph g) {
		distances = new ThreeDimensionalHashtable<String, Integer>(1);
		this.graph = g;
	}

	private String lowest(Set<String> list, final Map<String, Integer> scores) {
		ArrayList<String> slist = new ArrayList<String>(list.size());
		slist.addAll(list);
		Collections.sort(slist, new Comparator<String>() {
			public int compare(String arg0, String arg1) {
				Integer v0 = unnull(scores.get(arg0));
				Integer v1 = unnull(scores.get(arg1));
				return v0.compareTo(v1);
			}
		});
		return slist.get(0);
	}

	public List<String> getShortestPath(String source, String target) {
		Hashtable<String, String> came_from = new Hashtable<String, String>();
		Map<String, Integer> g_score = new Hashtable<String, Integer>();
		Map<String, Integer> f_score = new TreeMap<String, Integer>();
		Map<String, Integer> h_score = new TreeMap<String, Integer>();
		Set<String> closedlist = new HashSet<String>();
		Set<String> openlist = new HashSet<String>();
		openlist.add(source);
		g_score.put(source, 0);
		h_score.put(source, distances.get(source, target));
		f_score.put(source, distances.get(source, target));
		while (!openlist.isEmpty()) {
			String x = lowest(openlist, f_score);
			if (x.equals(target)) {
				return resolve_path(came_from, source, target);

			}
			openlist.remove(x);
			closedlist.add(x);
			for (String y : graph.getSuccessors(x)) {
				Integer est = 1;
				if (closedlist.contains(y))
					continue;
				Integer tentative_g_score = g_score.get(x) + 1;
				boolean tentative_is_better = false;
				if (!openlist.contains(y)) {
					openlist.add(y);
					est = distances.get(x, y);
					h_score.put(y, est);
					tentative_is_better = true;
				} else if (tentative_g_score < g_score.get(y)) {
					tentative_is_better = true;
					est = h_score.get(y);
				}
				if (tentative_is_better) {
					came_from.put(y, x);
					g_score.put(y, tentative_g_score);
					f_score.put(y, tentative_g_score + est);
					h_score.put(y, distances.get(y, target));
				}
			}
		}
		LinkedList<String> l = new LinkedList<String>();
		l.add(source);
		setDistance(source, target, Integer.MAX_VALUE);
		return l;
	}

	private int unnull(Integer integer) {
		if (integer == null)
			return 1;
		else
			return integer;
	}

	public int getCacheSize() {
		return distances.size();
	}

	private List<String> resolve_path(Hashtable<String, String> came_from,
			String source, String target) {
		List<String> path = new LinkedList<String>();
		String cnode = target;
		path.add(target);
		while (!cnode.equals(source)) {
			String prev = came_from.get(cnode);
			path.add(0, prev);
			cnode = prev;
		}
		int length = path.size() - 1;
		setDistance(source, target, length);
		return path;
	}

	private void setDistance(String source, String target, int length) {
		distances.put(source, target, length);
		if (!graph.isDirected())
			distances.put(target, source, length);
	}

	public static void unitTest() {
		Graph g = new Graph(true);
		System.out.println("Graph is directed? " + g.isDirected());
		g.addEdge("0", "1");
		g.addEdge("1", "2");
		g.addEdge("2", "3");
		g.addEdge("3", "4");
		g.addEdge("3", "5");
		g.addEdge("4", "8");
		g.addEdge("5", "6");
		g.addEdge("6", "7");
		g.addEdge("7", "8");
		g.addEdge("4", "8");
		g.addEdge("9", "3");
		g.addEdge("8", "9");

		// Unconnected part
		g.addEdge("10", "11");
		g.addEdge("11", "12");
		g.addEdge("12", "13");

		AStar as = new AStar(g);

		System.out.println("0->4 =>" + as.getShortestPath("0", "4"));
		System.out.println("0->5 =>" + as.getShortestPath("0", "5"));
		System.out.println("0->6 =>" + as.getShortestPath("0", "6"));
		System.out.println("0->8 =>" + as.getShortestPath("0", "8"));
		System.out.println("0->9 =>" + as.getShortestPath("0", "9"));
		System.out.println("0->7 =>" + as.getShortestPath("0", "7"));
		System.out.println("0->10 =>" + as.getShortestPath("0", "10"));
		System.out.println("13->10 =>" + as.getShortestPath("13", "10"));
	}

	public static void main(String[] args) {
		unitTest();
	}
}
