/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import services.dss.profile.ResourceProfile;
import support.Parameters;

/**
 * The Standard Workload Format
 * 
 * http://www.cs.huji.ac.il/labs/parallel/workload/swf.html
 * 
 * 
   0.  Job Number -- a counter field, starting from 1.
   1. Submit Time -- in seconds. The earliest time the log refers to is zero, and is the submittal time the of the first job. The lines in the log are sorted by ascending submittal times. It makes sense for jobs to also be numbered in this order.
   2. Wait Time -- in seconds. The difference between the job's submit time and the time at which it actually began to run. Naturally, this is only relevant to real logs, not to models.
   3. Run Time -- in seconds. The wall clock time the job was running (end time minus start time).
      We decided to use ``wait time'' and ``run time'' instead of the equivalent ``start time'' and ``end time'' because they are directly attributable to the scheduler and application, and are more suitable for models where only the run time is relevant.
      Note that when values are rounded to an integral number of seconds (as often happens in logs) a run time of 0 is possible and means the job ran for less than 0.5 seconds. On the other hand it is permissable to use floating point values for time fields.
   4. Number of Allocated Processors -- an integer. In most cases this is also the number of processors the job uses; if the job does not use all of them, we typically don't know about it.
   5. Average CPU Time Used -- both user and system, in seconds. This is the average over all processors of the CPU time used, and may therefore be smaller than the wall clock runtime. If a log contains the total CPU time used by all the processors, it is divided by the number of allocated processors to derive the average.
   6. Used Memory -- in kilobytes. This is again the average per processor.
   7. Requested Number of Processors.
   8. Requested Time. This can be either runtime (measured in wallclock seconds), or average CPU time per processor (also in seconds) -- the exact meaning is determined by a header comment. In many logs this field is used for the user runtime estimate (or upper bound) used in backfilling. If a log contains a request for total CPU time, it is divided by the number of requested processors.
   9. Requested Memory (again kilobytes per processor).
  10. Status 1 if the job was completed, 0 if it failed, and 5 if cancelled. If information about chekcpointing or swapping is included, other values are also possible. See usage note below. This field is meaningless for models, so would be -1.
  11. User ID -- a natural number, between one and the number of different users.
  12. Group ID -- a natural number, between one and the number of different groups. Some systems control resource usage by groups rather than by individual users.
  13. Executable (Application) Number -- a natural number, between one and the number of different applications appearing in the workload. in some logs, this might represent a script file used to run jobs rather than the executable directly; this should be noted in a header comment.
  14. Queue Number -- a natural number, between one and the number of different queues in the system. The nature of the system's queues should be explained in a header comment. This field is where batch and interactive jobs should be differentiated: we suggest the convention of denoting interactive jobs by 0.
  15. Partition Number -- a natural number, between one and the number of different partitions in the systems. The nature of the system's partitions should be explained in a header comment. For example, it is possible to use partition numbers to identify which machine in a cluster was used.
  16. Preceding Job Number -- this is the number of a previous job in the workload, such that the current job can only start after the termination of this preceding job. Together with the next field, this allows the workload to include feedback as described below.
  17. Think Time from Preceding Job -- this is the number of seconds that should elapse between the termination of the preceding job and the submittal of this one. 
 * @author attila
 *
 */

public class SWFLoader {
	
	private List<ResourceProfile> tasks;
	private int index = 0;
	
	public SWFLoader(String file) {
		tasks = load(file);
	}
	
	private List<ResourceProfile> load(String file) {
		String line = null;
		BufferedReader br = null;
		List<ResourceProfile> profiles = new ArrayList<ResourceProfile>();
		try {
			br = new BufferedReader(new FileReader(file));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}

		try {
			while ((line = br.readLine()) != null) {
				line = line.trim();
				if (line.charAt(0) == ';') {
					// This is data we don't care for
				} else {
					StringTokenizer st = new StringTokenizer(line);
					Long[] data = new Long[18];
					for (int i = 0; i < 18; i++) {
						String token = st.nextToken().trim();
						data[i] = Long.parseLong(token);
					}
					ResourceProfile p = new ResourceProfile();
					p.id = data[0].toString();
					p.pool = "n/a";
					p.createdAt = (data[1] * 1000) / Parameters.SWF_SCALE;
					p.actualRunningTime = (data[3] * 1000)  / Parameters.SWF_SCALE;
					p.setEstimatedRunningTime((data[8]* 1000)  / Parameters.SWF_SCALE);
					p.memory = data[9];
					p.actualCosts = new ArrayList<Long>();
					p.promisedCosts = new ArrayList<Long>();
					
					p.delays = new ArrayList<Long>();
					profiles.add(p);
				}
			}
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return profiles;
	}
	
	public ResourceProfile getJob(long timems) {
		if (index >= tasks.size()) return null;
		ResourceProfile p = tasks.get(index);
		if ((p.createdAt) <= timems) {
			index++;
			return p;
		} else {
			return null;
		}
	}

}
