/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class WeightedAStar {

	/*
	 * A*, shortest path Adapted from the Wikipedia, The Free Encyclopedia
	 * (http://en.wikipedia.org/wiki/A*_search_algorithm)
	 */

	class ThreeDimensionalHashMap<K, V> {

		private HashMap<K, HashMap<K, V>> data;
		private V nullvalue;

		public ThreeDimensionalHashMap(V nullvalue) {
			data = new HashMap<K, HashMap<K, V>>(800000);
			this.nullvalue = nullvalue;
		}

		public void put(K key1, K key2, V value) {
			if (data.containsKey(key1)) {
				data.get(key1).put(key2, value);
			} else {
				HashMap<K, V> entry = new HashMap<K, V>(80000);
				entry.put(key2, value);
				data.put(key1, entry);
			}
		}

		public V get(K key1, K key2) {
			if (data.containsKey(key1)) {
				HashMap<K, V> entry = data.get(key1);
				if (entry.containsKey(key2))
					return entry.get(key2);
			}
			return nullvalue;
		}

		public int size() {
			return data.size();
		}
		
		public void clear() {
			data.clear();
		}

	}

	private ThreeDimensionalHashMap<String, Integer> distances;
	private WeightedGraph graph;

	public WeightedAStar(WeightedGraph g) {
		distances = new ThreeDimensionalHashMap<String, Integer>(1);
		this.graph = g;
	}
	
	public void clear() {
		distances.clear();
	}

	private String lowest(Set<String> list, final Map<String, Integer> scores) {
		ArrayList<String> slist = new ArrayList<String>(list.size());
		slist.addAll(list);
		Collections.sort(slist, new Comparator<String>() {
			public int compare(String arg0, String arg1) {
				Integer v0 = unnull(scores.get(arg0));
				Integer v1 = unnull(scores.get(arg1));
				return v0.compareTo(v1);
			}
		});
		return slist.get(0);
	}

	// For the RTT we are not interested in the path, only the weight sum
	
	public List<String> getShortestPath(String source, String target) {
		HashMap<String, String> came_from = new HashMap<String, String>();
		Map<String, Integer> g_score = new HashMap<String, Integer>(8000);
		Map<String, Integer> f_score = new HashMap<String, Integer>(8000);
		Map<String, Integer> h_score = new HashMap<String, Integer>(8000);
		Set<String> closedlist = new HashSet<String>();
		Set<String> openlist = new HashSet<String>();
		openlist.add(source);
		g_score.put(source, 0);
		h_score.put(source, distances.get(source, target));
		f_score.put(source, distances.get(source, target));
		while (!openlist.isEmpty()) {
			String x = lowest(openlist, f_score);
			if (x.equals(target)) {
				return resolve_path(came_from, source, target);
			}
			openlist.remove(x);
			closedlist.add(x);
			for (String y : graph.getSuccessors(x)) {
				Integer est = 1;
				if (closedlist.contains(y))
					continue;
				Integer tentative_g_score = g_score.get(x) + graph.getWeight(x, y).intValue();
				boolean tentative_is_better = false;
				if (!openlist.contains(y)) {
					openlist.add(y);
					est = distances.get(x, y);
					h_score.put(y, est);
					tentative_is_better = true;
				} else if (tentative_g_score < g_score.get(y)) {
					tentative_is_better = true;
					est = h_score.get(y);
				}
				if (tentative_is_better) {
					came_from.put(y, x);
					g_score.put(y, tentative_g_score);
					h_score.put(y, distances.get(y, target));
					f_score.put(y, tentative_g_score + est);
				}
			}
		}
		LinkedList<String> l = new LinkedList<String>();
		l.add(source);
		setDistance(source, target, Integer.MAX_VALUE);
		return l;
	}

	private int unnull(Integer integer) {
		if (integer == null)
			return 1;
		else
			return integer;
	}

	public int getCacheSize() {
		return distances.size();
	}

	private List<String> resolve_path(HashMap<String, String> came_from,
			String source, String target) {
		List<String> path = new LinkedList<String>();
		String cnode = target;
		path.add(target);
		int distance = 0;
		while (!cnode.equals(source)) {
			String prev = came_from.get(cnode);
			distance += graph.getWeight(cnode, prev);
			path.add(0, prev);
			cnode = prev;
		}
		//int length = path.size() - 1;
		path.add(0,Integer.toString(distance));
		setDistance(source, target, distance);
		return path;
	}

	private void setDistance(String source, String target, int length) {
		distances.put(source, target, length);
		if (!graph.isDirected())
			distances.put(target, source, length);
	}
	
	public long getDistance(String source, String target) {
		return distances.get(source, target);
	}

}
